import justpy as jp
import asyncio
import matplotlib.pyplot as plt

import logging

logger = logging.getLogger("qolab.gui.web")

button_classes = (
    "bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full"
)
label_div_classes = "py-1 px-2"
checkbox_classed = label_div_classes
labelnames_classes = "font-bold"
panel_div_classes = "space-x-4 border"
controls_div_classes = "flex flex-wrap space-x-4 p-1 border"
controls_group_classes = "flex flex-wrap space-x-4 p-1"
dict_classes = "px-2 border-2"
dict_components_classes = "flex flex-wrap border-2"
input_classes = (
    "m-2 bg-gray-200"
    + " border-2 border-gray-200"  # noqa: W503 disables flake8 warning
    + " rounded w-20 text-gray-700"  # noqa: W503
    + " focus:outline-none focus:bg-white focus:border-purple-500"  # noqa: W503
)


class QOLPushButton(jp.Button):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_classes(
            "bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full"
        )


class QOLPushButtonNoUndo(QOLPushButton):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_classes(
            "bg-red-500 hover:bg-red-700 text-white font-bold py-2 px-4 rounded-full"
        )


class QOLParamHeader(jp.Div):
    def __init__(self, label="NameNotSet", value=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.label = label
        root.set_classes(label_div_classes)
        root.data[label] = None
        jp.Span(text=f"{label}: ", classes=labelnames_classes, a=root)
        self.setValue(value)

    def getValue(self):
        return self.data[self.label]

    def setValue(self, val):
        # print(f'setting {self.label} to {val}')
        self.data[self.label] = val


class QOLParamReadOnly(QOLParamHeader):
    """Read Only from the web point of view"""

    def __init__(self, label="NameNotSet", **kwargs):
        super().__init__(label=label, **kwargs)
        root = self
        jp.Span(model=[root, label], a=root)


class QOLParamReadWrite(QOLParamHeader):
    def __init__(self, label="NameNotSet", **kwargs):
        super().__init__(label=label, **kwargs)
        root = self
        self.input = jp.InputChangeOnly(
            classes=input_classes, model=[root, label], a=root, spellcheck="false"
        )


class QOLCheckbox(jp.Label):
    def __init__(self, label="NameNotSet", **kwargs):
        super().__init__(**kwargs)
        root = self
        root.data["checked"] = False
        root.label = label
        root.set_classes(checkbox_classed)
        jp.Input(
            type="checkbox", model=[root, "checked"], classes="form-checkbox", a=root
        )
        jp.Span(text=label, a=root)

    def getValue(self):
        return self.data["checked"]

    def setValue(self, val):
        self.data["checked"] = val


class QOLTimeLog(jp.Div):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.traces = None
        # log must have 'plot' and 'clear_data' methods
        root.set_classes(panel_div_classes)
        dcontrols = jp.Div(a=root)
        dcontrols.set_classes(controls_div_classes)
        QOLPushButtonNoUndo(a=dcontrols, text="Clear log", click=self._clear_data)
        QOLPushButton(a=dcontrols, text="Replot", click=self._replot)
        self.update_interval = QOLParamReadOnly(
            label="UpdateInterval", value=5, a=dcontrols
        )
        self.save_controls = QOLSaveControls(
            component_with_data=self.traces, a=dcontrols
        )

        self.chart = jp.Matplotlib(a=root)
        self.start_task()

    def setTraces(self, component_with_data=None):
        self.traces = component_with_data
        self.save_controls.component_with_data = component_with_data

    def start_task(self):
        jp.run_task(self.update_loop())

    def stop_tasks(self):
        self.save_controls.stop_tasks()
        pass

    async def _clear_data(self, msg):
        self.clear_data()
        await self.update()

    def clear_data(self):
        traces = self.traces
        if traces is not None:
            traces.clear_data()
            self.plot()

    async def _replot(self, msg):
        self.plot()
        await self.update()

    async def update_loop(self, update_interval=None):
        if update_interval is not None:
            self.update_interval.setValue(update_interval)
        while True:
            self.plot()
            try:
                strwtime = self.update_interval.getValue()
                wtime = float(strwtime)
            except ValueError:
                wtime = 2
                print(f"Warning cannot convert {strwtime} setting to {wtime}")
            await self.update()
            await asyncio.sleep(wtime)

    def plot(self):
        traces = self.traces
        f = plt.figure(figsize=(8, 6), tight_layout=True)
        if traces is None:
            plt.title("Log data is unavailable")
        else:
            traces.plot()
        self.chart.set_figure(f)
        plt.close(f)


class QOLSaveControls(jp.Div):
    def __init__(self, component_with_data=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.set_classes(controls_group_classes)

        root.component_with_data = component_with_data
        root.getNextDataFile = None
        root.bsave = QOLPushButton(a=root, text="Save", name="Save", click=self._save)
        root.bnext_file = QOLPushButton(
            a=root, text="Next file", name="NextFile", click=self._next_file
        )
        self.autosave_flag = QOLCheckbox(label="autosave", a=root)
        self.autosave_interval = QOLParamReadOnly(
            label="AutoSaveInterval", value=10, a=root
        )
        self.file_name = QOLParamReadOnly(label="FileName", a=root)
        self.start_task()

    def start_task(self):
        jp.run_task(self.autosave_loop())

    def stop_tasks(self):
        pass

    async def autosave_loop(self):
        while True:
            if self.autosave_flag.getValue():
                self.save()
            await asyncio.sleep(10)

    async def _save(self, msg):
        self.save()

    def save(self):
        fname = self.file_name.getValue()
        if fname is None:
            return
        if self.component_with_data is None:
            return
        logger.debug(f"saving to {fname}")
        self.component_with_data.save(fname)

    async def _next_file(self, msg):
        self.next_file()
        await self.update()

    def next_file(self):
        if self.getNextDataFile is not None:
            fname = self.getNextDataFile()
            logger.info(f"Data will be saved to {fname}")
            self.file_name.setValue(fname)
            if self.component_with_data is not None:
                self.component_with_data.clear_last_saved_pos()


class QOLDictionary(jp.Div):
    def __init__(self, container=None, **kwargs):
        super().__init__(**kwargs)
        root = self
        root.set_classes(dict_classes)
        self.dlabel = jp.Div(a=root, classes="bg-gray-100")
        self.dlabel.on("click", self._toggle_show)
        root.slabel = jp.Span(text=self.name, a=root.dlabel)
        root.slabel.set_classes(labelnames_classes)
        self.c = jp.Div(a=root)
        root.c.set_classes(dict_components_classes)
        root.container = container
        self.display_container_dictionary()

    async def _toggle_show(self, msg):
        self.c.show = not self.c.show

    def display_container_dictionary(self):
        self.c.delete_components()
        if self.container is None:
            return
        for k, v in self.container.items():
            if not isinstance(v, dict):
                QOLParamReadOnly(label=k, value=v, a=self.c)
            else:
                QOLDictionary(container=v, name=k, a=self.c)


def gui_test():
    return wp


if __name__ == "__main__":
    wp = jp.WebPage(delete_flag=False)
    rw = QOLParamReadWrite(label="ReadWriteParam", a=wp)
    rw.setValue(12345)

    log = QOLTimeLog(a=wp)
    log.save_controls.getNextDataFile = lambda: "data.dat"

    def test(self, msg):
        print(rw.getValue())

    QOLPushButtonNoUndo(text="Danger", a=wp, onclick=test)
    # sc = QOLSaveControls(a=wp)
    d = {}
    d["a"] = "astring"
    d["n"] = 7
    d["d"] = {}
    for i in range(20):
        d["d"][i] = i
    QOLDictionary(a=wp, name="d dictionary", container=d)

    jp.justpy(gui_test)



import numpy as np
from scipy.fft import rfft, rfftfreq


def spectrum(t, y):
    """
    Calculate amplitude spectrum of real value signal.

    Unlike ``scipy.fft.rfft()``, preserves amplitude (A) of a coherent signal
    ``A*sin(f*t)`` independent of sampling rate and time interval.
    I.e. at frequency 'f' we will get 'A'.

    It also strips away zero frequency component.

    Returns
    -------
    array of frequencies and corresponding amplitudes
    """
    N = t.size
    dt = np.mean(np.diff(t))
    freq = rfftfreq(N, dt)
    # y= y - np.mean(y)
    yf = rfft(y)
    yf *= 2 / N  # produce normalized amplitudes
    return freq[1:], yf[1:]  # strip of boring freq=0


def noise_density_spectrum(t, y):
    """
    Calculate noise amplitude spectral density (ASD) spectrum.

    The end results has units of [units of y]/sqrt(Hz).
    I.e. it calculates sqrt(PSD) where PSD is power spectrum density.
    Preserves the density independent of sampling rate and time interval.

    It also strips away zero frequency component, since it uses ``spectrum``

    Returns
    -------
    reduced array of frequencies and corresponding ASDs
    """
    freq, yf = spectrum(t, y)
    yf = yf * np.sqrt(t[-1] - t[0])  # scales with 1/sqrt(RBW)
    return freq, yf


def noise_spectrum_smooth(fr, Ampl, Nbins=100):
    """
    Smooth amplitude spectral density (ASD) spectrum.

    Could be thought as a running average with logarithmic spacing, so high
    frequency components average  bins with more "hits" and thus smoothed the most.

    Since we assume the input spectrum of the nose (ASD),
    we do power average (rmsq on amplitudes).

    Assumes that the input frequencies are in positive and equidistant set.
    Also assumes that frequencies do not contain 0.

    Parametes
    ---------
    fr : list or array
        array of frequencies
    Ampl : list or array
        array of corresponding noise amplitudes ASD
    Nbin : integer
        number of the bins (default is 100) in which frequencies are split

    Returns
    -------
    reduced array of frequencies and corresponding ASDs
    """

    frEdges = np.logspace(np.log10(fr[0]), np.log10(fr[-1]), Nbins)
    frCenter = np.zeros(frEdges.size - 1)
    power = np.zeros(frEdges.size - 1)
    for i, (frStart, frEnd) in enumerate(zip(frEdges[0:-1], frEdges[1:])):
        # print (f"{i=} {frStart=} {frEnd}")
        ind = (frStart <= fr) & (fr <= frEnd)
        frCenter[i] = np.mean(fr[ind])
        power[i] = np.mean(np.power(np.abs(Ampl[ind]), 2))
    ind = np.logical_not(np.isnan(frCenter))
    frCenter = frCenter[ind]
    power = power[ind]
    # print(f'{frCenter=} {power=}')
    return frCenter, np.sqrt(power)



"""
Provide basic class to build an operational scope

Created by Eugeniy E. Mikhailov 2021/11/29
"""

import numpy as np
from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument
from qolab.data.trace import TraceSetSameX, TraceXY
import time
import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def calcSparsingAndNumPoints(availableNpnts=None, maxRequiredPoints=None):
    """Calculate sparcing and number of sparced points.

    Parameters
    ----------
    availableNpnts: int or None (throws error)
        Number of available points. If set to None exit with error
    maxRequiredPoints: int or None (throws error)
        number of requested points after decimation.
        If availableNpnts< maxRequiredPoints*2,
        decimation is impossible and we will get up to factor of 2 more
        than requested.

    Return
    ------
    (sparsing, Npnts, availableNpnts, maxRequiredPoints)
    """
    if availableNpnts is None:
        raise ValueError("Invalid availableNpnts value, must be int.")
    if maxRequiredPoints is None:
        raise ValueError("Invalid maxRequiredPoints value, must be int.")

    if availableNpnts <= maxRequiredPoints * 2:
        Npnts = availableNpnts
        sparsing = 1
    else:
        sparsing = int(np.floor(availableNpnts / maxRequiredPoints))
        Npnts = int(np.floor(availableNpnts / sparsing))
    return (sparsing, Npnts, availableNpnts, maxRequiredPoints)


class Scope(BasicInstrument):
    """Minimal class to implement a scope.

    Intended to be used as a parent for hardware aware scopes.

    Provide a minimal set of methods to be implemented by a scope.
    """

    vertDivOnScreen = 8  # should be adjusted in hardware aware class
    horizDivOnScreen = 8  # should be adjusted in hardware aware class
    minVoltsPerDiv = 0.001  # should be adjusted in hardware aware class

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "Scope"
        self.config["Device model"] = "Generic Scope Without Hardware interface"
        self.config["FnamePrefix"] = "scope"
        self.numberOfChannels = 0
        # deviceProperties must have 'get' and preferably 'set' methods available,
        # i.e. 'SampleRate' needs getSampleRate() and love to have setSampleRate(value)
        # they will be used to obtain config and set device according to it
        self.deviceProperties.update(
            {"SampleRate", "TimePerDiv", "TrigDelay", "TriggerMode", "Roll", "Run"}
        )
        # same is applied to channelProperties
        # but we need setter/getter with channel number
        # i.e.  VoltsPerDiv must provide
        # getChanVoltsPerDiv(chNum) and setSampleRate(chNum, value)
        self.channelProperties = {
            "VoltsPerDiv",
            "VoltageOffset",
        }

    def getWaveform(
        self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True
    ):
        """
        Get scope channel waveform where X axis is index and not time.

        Waveform MUST HAVE entry in config dictionary such as
        ``wafeform.config["tags"]["rawChanConfig"]``
        with items helping to calculate time trace
        (Usually ``SampleRate``, ``Npnts``, ``sparsing``).

        For decimate use see ``getRawWaveform``.

        In short decimate=True is slower but more precise.

        """
        raise NotImplementedError("getWaveform function is not implemented")

    def getTimeTrace(self, rawChanCfg):
        """Constructs time trace from properties in ``rawChanCfg`` dictionary."""
        raise NotImplementedError("getTimeTrace function is not implemented")

    def getTrace(
        self, 
        chNum, 
        availableNpnts=None, 
        maxRequiredPoints=None, 
        decimate=True,
        fmath=False
    ):
        """Get scope trace with time axis set."""
        old_run_status = self.getRun()
        if old_run_status:  # avoid unnecessary status change
            self.setRun(False)  # stop if currently running
            self._waitUntillStop()
        # to get synchronous channels
        wfVoltage = self.getWaveform(
            chNum,
            availableNpnts=availableNpnts,
            maxRequiredPoints=maxRequiredPoints,
            decimate=decimate,
            fmath=fmath
        )
        rawChanCfg = wfVoltage.config["tags"]["rawChanConfig"]
        t = self.getTimeTrace(rawChanCfg)
        if fmath:
            tr = TraceXY(f"ChF{chNum}")
        else:
            tr = TraceXY(f"Ch{chNum}")
        tr.x = t
        tr.y = wfVoltage
        # restore scope to the before acquisition mode
        if old_run_status:  # avoid unnecessary status change
            self.setRun(old_run_status)  # start running if it was old run state
        return tr

    def getTriggerMode(self):
        # we expect NORM, AUTO, SINGLE
        raise NotImplementedError("getTriggerMode function is not implemented")

    def setTriggerMode(self, mode):
        # we expect NORM, AUTO, SINGLE
        raise NotImplementedError("setTriggerMode function is not implemented")

    def getRun(self):
        """Is acquisition running or stopped."""
        raise NotImplementedError("getRun function is not implemented")

    def setRun(self, val):
        """Either enable run or stop the acquisition."""
        raise NotImplementedError("setRun function is not implemented")

    def _waitUntillStop(self, timeout=1):
        """Wait until scope in the stop state.

        Just because we ask for a scope to stop, does not mean
        that it is stopped. It can still wait for a trigger or untill
        the time span is filled.

        Parameter
        ---------
        timeout : float
            timeout in seconds, default is 1 second
        """
        starttime = time.time()
        deadline = starttime + timeout
        while time.time() < deadline:
            if self.getRun():
                time.sleep(0.010)
            else:
                logger.debug(f"Scope stopped within {time.time()-starttime} seconds.")
                return
        logger.warning(
            f"Scope did not reach STOP state within {timeout=} sec, try to increase it."
        )

    def getAllTraces(self, availableNpnts=None, maxRequiredPoints=None, decimate=True):
        allTraces = TraceSetSameX("scope traces")
        allTraces.config["tags"]["DAQ"] = self.getConfig()
        old_run_status = self.getRun()
        if old_run_status:  # avoid unnecessary status change
            self.setRun(False)  # stop if currently running
            self._waitUntillStop()
        # to get synchronous channels
        for chNum in range(1, self.numberOfChannels + 1):
            allTraces.addTrace(
                self.getTrace(
                    chNum,
                    availableNpnts=availableNpnts,
                    maxRequiredPoints=maxRequiredPoints,
                    decimate=decimate,
                )
            )
        # restore scope to the before acquisition mode
        if old_run_status:  # avoid unnecessary status change
            self.setRun(old_run_status)  # start running if it was old run state
        return allTraces

    def restorePriorToFastGrab(self, chNum, old_config):
        """Restore relevant channel/scope settings prior to fast grab tune up."""
        logger.info("method restorePriorToFastGrab is not implemented")

    def switchToFastGrab(self, chNum):
        """Switch scope to the fastest mode for trace delivery.

        Intended to be used in chanAutoScale methods where speed
        transfer is more important than fidelity. It is good
        idea to implement function ``restorePriorToFastGrab(chan, old_config)``
        to be used in conjunction with ``switchToFastGrab``.

        To be fast, it should aim to decrease data transfer time,
        i.e. reduce number of transferred point.
        It also need to be interface dependent. GPIB and USB are known
        to be quite slow when compared to network connection.

        So try to be smart in hardware dependent classes.

        Return
        ------
        old_config : dictionary
            old config with settings necessary to restore initial state
        """

        logger.info("machinery for switchToFastGrab is not implemented")
        return self.getConfig()

    def chanAutoScale(self, chNum, margin=0.125, timeout=5):
        """Auto scale channel to fit signal on screen.

        Tunes Volts per division and Channel offset to fit signal
        on screen (vertically).

        Parameters
        ----------
        chNum : int
            Channel to auto scale
        margin: float
           How much extra space (margin) to have with respect to full screen.
           Default is 0.125 (i.e. 12.5%).
           Note that margin = 0.25 corresponds to 1 vertical division
           at top and bottom for 8 division scope.
        """

        old_config = self.switchToFastGrab(chNum)
        starttime = time.time()
        deadline = starttime + timeout
        timespan = self.getTimePerDiv() * self.horizDivOnScreen
        scaled_corectly = False
        self.setRun(True)

        while (not scaled_corectly) and (time.time() < deadline):
            time.sleep(
                timespan + 0.5
            )  # give enough time to acquire a trace and switch to Run
            tr = self.getTrace(chNum)
            vPerDiv = self.getChanVoltsPerDiv(chNum)
            offset = self.getChanVoltageOffset(chNum)
            v_range = vPerDiv * self.vertDivOnScreen
            v_max = v_range / 2 - offset
            v_min = -v_range / 2 - offset
            y = tr.y.values
            tr_max = y.max()
            tr_min = y.min()
            signal_range = tr_max - tr_min
            margin_t = (v_max - tr_max) / v_range
            margin_b = (tr_min - v_min) / v_range
            is_margin_t_good = (margin_t > margin * 0.5) and (margin_t < margin * 1.0)
            is_margin_b_good = (margin_b > margin * 0.5) and (margin_b < margin * 1.0)
            if (is_margin_b_good) and (is_margin_t_good):
                scaled_corectly = True
                break
            old_offset = offset
            old_vPerDiv = vPerDiv
            offset = -(tr_max + tr_min) / 2
            if min(margin_t, margin_b) < 0.01:  # too close to edge, zoom out
                vPerDiv *= 2
            elif (signal_range / v_range) < 0.2:  # signal is too small, zoom in
                vPerDiv /= 2
            else:
                vPerDiv = signal_range / (self.vertDivOnScreen * (1 - margin * 1.2))
            if vPerDiv < self.minVoltsPerDiv:
                vPerDiv = self.minVoltsPerDiv
            logger.debug(f"Auto Scaler Requesting {vPerDiv=} for {chNum=}.")
            self.setChanVoltsPerDiv(chNum, vPerDiv)
            logger.debug(f"Auto Scaler Requesting {offset=} for {chNum=}.")
            self.setChanVoltageOffset(chNum, offset)
            relOffsetChange = abs(old_offset - offset) / old_vPerDiv
            relVperDivChange = abs(old_vPerDiv - vPerDiv) / old_vPerDiv
            if (relOffsetChange < 0.5) and (relVperDivChange < 0.05):
                # we converge, there is no point to improve
                break
            scaled_corectly = False
        self.restorePriorToFastGrab(chNum, old_config)
        if time.time() > deadline:
            logger.warning(
                f"Scope did not make proper channel {chNum}"
                + f"scaling within {timeout=} sec."
            )

    def plot(self, **kwargs):
        allTraces = self.getAllTraces(**kwargs)
        allTraces.plot()

    def save(
        self,
        fname=None,
        item_format="e",
        availableNpnts=None,
        maxRequiredPoints=None,
        decimate=True,
        extension="dat",
    ):
        allTraces = self.getAllTraces(
            availableNpnts=availableNpnts,
            maxRequiredPoints=maxRequiredPoints,
            decimate=decimate,
        )
        allTraces.config["item_format"] = item_format
        if fname is None:
            fname = self.getNextDataFile(extension=extension)
        allTraces.save(fname)
        print(f"Data saved to: {fname}")
        return fname


class ScopeSCPI(SCPIinstr, Scope):
    """SCPI aware scope.

    Use as a parent for a hardware aware scope classes.

    Example
    -------

    >>> rm = pyvisa.ResourceManager()
    >>> ScopeSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Scope.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



from ._basic import DAQ
import ue9  # not availabe via pip, download from LabJack website

# https://support.labjack.com/docs/software-driver


class LabJackUE9(DAQ):
    """
    DAQ LabJack UE9 class.

    This is custom modification of the stock ue9 class
    provided by LabJack.
    """

    def __init__(
        self, *args, debug=False, autoOpen=True, ethernet=False, ipAddress=None, **kargs
    ):
        """
        LabJack UE9 can be contacted via TCP, use
        LabJackUE9(ethernet=True, ipAddress="192.168.1.209")
        """
        super().__init__(*args, **kargs)
        self.config["Device model"] = "LabJack UE9"
        self.AnalogInputsNum = 4
        self.AnalogOutputsNum = 2

        self.daq = ue9.UE9(
            debug=debug, autoOpen=autoOpen, ethernet=ethernet, ipAddress=ipAddress
        )

        # For applying the proper calibration to readings.
        c = self.daq.getCalibrationData()

        # by evmik
        # fixing missing slope for gain '0'
        c["AINSlopes"]["0"] = 0.0000775030

    def getAIN(self, chNum):
        """get analog input (AIN) voltage"""
        BipGain = 8
        Resolution = 12
        SettlingTime = 0
        # BipGain = 8 -> bipolar range (-5V, +5V) gain 1
        # UE9 default BipGain = 0 -> signal range (0V, +5V) gain 1
        # other BipGain could be:
        #   0 = Unipolar Gain 1, 1 = Unipolar Gain 2,
        #   2 = Unipolar Gain 4, 3 = Unipolar Gain 8,
        #   8 = Bipolar Gain 1
        return self.daq.getAIN(
            chNum, BipGain=BipGain, Resolution=Resolution, SettlingTime=SettlingTime
        )

    def setDAC(self, chNum, volts):
        """set digital to analog (DAC) output voltage"""
        if (chNum is None) or (volts is None):
            print("setOutputCh needs chNum and volts to be set")
            return 0
        bits = self.daq.voltageToDACBits(volts, dacNumber=chNum)
        # results are completely bogus for DAC settings in UE9
        self.daq.singleIO(IOType=5, Channel=chNum, DAC=bits)
        return volts

    def close(self):
        self.daq.close()


if __name__ == "__main__":
    daq = LabJackUE9()
    print("testing")
    print("------ Header start -------------")
    print(str.join("\n", daq.getHeader()))
    print("------ Header ends  -------------")



import logging
import justpy as jp
import pyvisa
import numpy as np
import qolab.tsdb as tsdb
import asyncio
import time
import qolab.gui.web as gui
from qolab.hardware import BasicInstrument
from qolab.hardware.scope import SDS1104X
from qolab.hardware.rf_generator import AgilentE8257D
from qolab.hardware.lockin import SRS_SR865A
from qolab.feedback import PID
from qolab.data import TraceSetSameX, TraceXY, Trace
from qolab.hardware.i_server import I800

# this should be done before justpy is called or log formatter does not work
logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)


tmplog = logging.getLogger("qolab.gui.web")
tmplog.setLevel(logging.INFO)
tmplog = logging.getLogger("qolab.tsdb")
tmplog.setLevel(logging.INFO)
logger = logging.getLogger("Magnetometer")
logger.setLevel(logging.INFO)


def getConfig(apparatus):
    config = apparatus.config.copy()
    ai = apparatus.instruments
    for n, i in ai.items():
        config[n] = i.getConfig()
    return config


def initLog(extra_tags={}):
    errorTrace = Trace("error")
    errorTrace.config["unit"] = "V"
    timeTrace = Trace("time")
    timeTrace.config["unit"] = "S"
    timeTrace.config["type"] = "timestamp"
    errorLog = TraceXY("error")
    errorLog.x = timeTrace
    errorLog.y = errorTrace

    freqTrace = Trace("frequency")
    freqTrace.config["unit"] = "Hz"
    freqLog = TraceXY("frequency")
    freqLog.x = timeTrace
    freqLog.y = freqTrace

    feedbackTrace = Trace("feedback")
    feedbackTrace.config["unit"] = "Hz"
    feedbackLog = TraceXY("feedback")
    feedbackLog.x = timeTrace
    feedbackLog.y = feedbackTrace

    eitTrace = Trace("eit")
    eitTrace.config["unit"] = "V"
    eitLog = TraceXY("eit")
    eitLog.x = timeTrace
    eitLog.y = eitTrace

    rfPoutTrace = Trace("rfPout")
    rfPoutTrace.config["unit"] = "dBm"
    rfPoutLog = TraceXY("rfPout")
    rfPoutLog.x = timeTrace
    rfPoutLog.y = rfPoutTrace

    cellTemperatureLog = TraceXY("cellTemperature")
    cellTemperatureLog.x = timeTrace
    cellTemperatureLog.y = Trace("cellTemperature")
    cellTemperatureLog.y.config["unit"] = "C"

    log = TraceSetSameX("timelog")
    log.addTrace(errorLog)
    log.addTrace(freqLog)
    log.addTrace(feedbackLog)
    log.addTrace(eitLog)
    log.addTrace(rfPoutLog)
    log.addTrace(cellTemperatureLog)
    log.config["tags"]["apparatus"] = getConfig(apparatus)
    log.config["tags"].update(extra_tags)
    return log


async def feedbackLoop(apparatus, nsteps):
    # while True:
    for i in range(0, nsteps):
        adjustRFandLog(apparatus)
        await asyncio.sleep(0.1)
    apparatus.runStatus = False


def adjustRFandLog(apparatus):
    timenow = time.time()

    ai = apparatus.instruments
    pid = ai["pid"]
    log = apparatus.gui_log.traces

    err, eit = apparatus.getErrAndEit()
    fdbck = pid.feedback(err)

    freq0 = apparatus.getRFreq()
    apparatus.setFeefback(fdbck)

    rfPout = apparatus.getRFAmplitude()

    apparatus.getBCurrent()
    # this automatically logs value to TSDB

    tCell = ai["cellTemperatureController"].getTemperature()

    log.addPointToTrace(timenow)
    log.addPointToTrace(err, "error")
    log.addPointToTrace(freq0, "frequency")
    log.addPointToTrace(fdbck, "feedback")
    log.addPointToTrace(rfPout, "rfPout")
    log.addPointToTrace(eit, "eit")
    log.addPointToTrace(tCell, "cellTemperature")


async def initial_lock_to_eit(apparatus, extra_tags={}):
    rfPout = apparatus.getRFAmplitude()
    apparatus.state = f"Initial lock RF power {rfPout} dBm"
    logger.info(apparatus.state)
    update_webpage(apparatus=apparatus)

    ai = apparatus.instruments
    apparatus.gui_log.setTraces(initLog(extra_tags=extra_tags))
    ai["pid"].reset()
    ai["pid"].setEnable(True)
    apparatus.runStatus = True
    await asyncio.gather(feedbackLoop(apparatus, nsteps=50))


async def calibratingLockin(apparatus, extra_tags={}):
    ai = apparatus.instruments
    rfPout = apparatus.getRFAmplitude()
    apparatus.state = f"Calibrating lockin response at RF power {rfPout} dBm"
    logger.info(apparatus.state)
    update_webpage(apparatus=apparatus)
    ai["pid"].setEnable(False)
    ai["pid"].reset()
    apparatus.gui_log.setTraces(initLog(extra_tags=extra_tags))
    await asyncio.gather(feedbackLoop(apparatus, nsteps=20))

    fr0 = ai["rfgen"].getFreqFixed()
    df = 10
    ai["pid"].setEnable(False)
    ai["pid"].reset()
    apparatus.setRFreq(fr0 + df)
    await asyncio.gather(feedbackLoop(apparatus, nsteps=20))
    apparatus.setRFreq(fr0)

    log = apparatus.gui_log.traces
    trE = log.getTrace("error")

    e1 = np.mean(trE.y.values[0:20])
    e2 = np.mean(trE.y.values[20:])
    dE = e2 - e1
    slope = dE / df
    apparatus.setLockinSlope(slope)

    logger.info(f"dE = {dE}")
    logger.info(f"lockin error signal slope = {slope} V/Hz")
    ai["error_signal_response_to_eit_detuning"].conversion_factor = float(slope)

    return float(slope)


async def responseToChangeOfBfieldControlVoltage(apparatus, extra_tags={}):
    ai = apparatus.instruments
    rfPout = apparatus.getRFAmplitude()
    apparatus.state = (
        f"Calibrating lockin response to change of Bfield control voltage, {rfPout} dBm"
    )
    logger.info(apparatus.state)
    update_webpage(apparatus=apparatus)
    ai["pid"].setEnable(True)
    ai["pid"].reset()
    apparatus.gui_log.setTraces(initLog(extra_tags=extra_tags))
    dV = 0.01
    ai["lockin"].AuxOut1 = dV

    # initial lock
    ai["pid"].setEnable(True)
    ai["pid"].reset()
    await asyncio.gather(feedbackLoop(apparatus, nsteps=20))

    # this is for the money lock
    await asyncio.gather(feedbackLoop(apparatus, nsteps=30))

    fr0 = ai["rfgen"].getFreqFixed()

    # this is for the money lock
    ai["lockin"].AuxOut1 = 0
    ai["pid"].setEnable(True)
    ai["pid"].reset()
    await asyncio.gather(feedbackLoop(apparatus, nsteps=30))
    fr1 = ai["rfgen"].getFreqFixed()

    # log = apparatus.gui_log.traces
    # trE = log.getTrace("error")

    df = fr0 - fr1
    slope = df / dV

    logger.info(f"df = {df}")
    logger.info(f"response to magnetic control voltage = {slope} Hz/V")
    ai["B_control_voltage_to_eit_detuning"].conversion_factor = float(slope)

    # relock to default state
    ai["lockin"].AuxOut1 = 0
    ai["pid"].setEnable(True)
    ai["pid"].reset()
    await asyncio.gather(feedbackLoop(apparatus, nsteps=20))

    return float(slope)


async def record_magnetometer_noise(apparatus):
    ai = apparatus.instruments
    rfPout = apparatus.getRFAmplitude()
    ai["pid"].setEnable(False)

    apparatus.state = f"Record magnetometer noise,  RF power {rfPout} dBm"
    logger.info(apparatus.state)

    scope = ai["scope.monitor"]
    old_config = scope.getConfig()
    tperdiv = 1
    scope.setTimePerDiv(tperdiv)
    scope.setRoll(True)

    await asyncio.sleep(tperdiv * 14 + 1)
    errorTr = scope.getTrace(1, maxRequiredPoints=10000)
    errorTr.config["label"] = "error vs time"
    errorTr.y.config["label"] = "error"
    log = errorTr
    log.config["tags"]["apparatus"] = getConfig(apparatus)
    magNoise = BasicInstrument(device_nickname=".".join(["magnetometer", "noise"]))
    magNoise.config["FnamePrefix"] = "magnetometer_noise"
    magNoise.config["SavePath"] = apparatus.config["SavePath"]
    fname = magNoise.getNextDataFile()
    log.save(fname)
    logger.info(f"Magnetometer noise file: {fname}")

    apparatus.gui_log.setTraces(log)
    update_webpage(apparatus=apparatus)

    logger.info("Restoring scope settings")
    scope.setConfig(old_config)


async def longTermLock(apparatus, extra_tags={}):
    ai = apparatus.instruments
    rfPout = apparatus.getRFAmplitude()
    apparatus.state = f"Long term lock RF power {rfPout} dBm"
    logger.info(apparatus.state)

    ai = apparatus.instruments
    rfPout = apparatus.getRFAmplitude()
    apparatus.runStatus = True
    ai["pid"].setEnable(True)

    apparatus.gui_log.setTraces(initLog(extra_tags=extra_tags))
    update_webpage(apparatus=apparatus)

    res = await asyncio.gather(feedbackLoop(apparatus, nsteps=10000000))
    return res


async def sweepRFPower(apparatus, extra_tags={}, rfPowerList=[]):
    apparatus.gui_log.setTraces(initLog(extra_tags=extra_tags))
    ai = apparatus.instruments
    # for p in np.linspace(-10,10, 110):
    ai["pid"].setEnable(True)
    for p in rfPowerList:
        ai["rfgen"].setRFAmplitude(p)
        rfPout = ai["rfgen"].getRFAmplitude()
        apparatus.state = f"lock with RF power {rfPout} dBm"
        logger.info(apparatus.state)
        update_webpage(apparatus=apparatus)
        ai["pid"].reset()
        apparatus.runStatus = True
        res = await asyncio.gather(feedbackLoop(apparatus, nsteps=50))
    return res


freqZero = 6_834_686_400
# freqDeltaMp2 = 6_835_385_570; # T=83.7C
freqDeltaMp2 = 6_835_393_258
# T=60C
# freqDeltaMp2 = 6_835_396_000; # T=44.6C
dfB = freqDeltaMp2 - freqZero
freqDeltaMm2 = freqZero - dfB


async def main():
    app_nickname = apparatus.config["DeviceNickname"]

    task_wp_update_loop = asyncio.create_task(update_webpage_loop(update_interval=1))

    apparatus.gui_log = gui.QOLTimeLog(a=wp)
    apparatus.gui_log.save_controls.getNextDataFile = apparatus.getNextDataFile

    logger.info("Accessing hardware")
    rm = pyvisa.ResourceManager()
    instr_scope = rm.open_resource("TCPIP::192.168.0.61::INSTR")
    scope_fdbk = SDS1104X(
        instr_scope,
        device_nickname=".".join([app_nickname, "scope.feedback"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_scope = rm.open_resource("TCPIP::192.168.0.62::INSTR")
    scope_mon = SDS1104X(
        instr_scope,
        device_nickname=".".join([app_nickname, "scope.monitor"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_rfgen = rm.open_resource("TCPIP::192.168.0.114::INSTR")
    rfgen = AgilentE8257D(
        instr_rfgen,
        device_nickname=".".join([app_nickname, "rfgen"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_lockin = rm.open_resource("TCPIP::192.168.0.51::INSTR")
    lockin = SRS_SR865A(
        instr_lockin,
        device_nickname=".".join([app_nickname, "lockin"]),
        tsdb_ingester=tsdb_ingester,
    )
    # print('------ Header start -------------')
    # print(str.join('\n', scope.getHeader()))
    # print(str.join('\n', rfgen.getHeader()))
    # print('------ Header ends  -------------')
    # ch1  = scope.getTrace(1)
    # traces = scope.getAllTraces()
    # pid = PID(100,400,0, sign=-1); # good for dm=-2 resonance
    pid = PID(
        5,
        20,
        0,
        sign=-1,
        device_nickname=".".join([app_nickname, "pid"]),
        tsdb_ingester=tsdb_ingester,
    )

    apparatus.instruments = {}
    ai = apparatus.instruments
    ai["rfgen"] = rfgen
    ai["lockin"] = lockin
    ai["pid"] = pid
    ai["cellTemperatureController"] = I800(
        device_nickname=".".join([app_nickname, "cellTemperatureController"]),
        tsdb_ingester=tsdb_ingester,
    )
    ai["scope.feedback"] = scope_fdbk
    scope_fdbk.setRoll(False)
    scope_fdbk.setRun(True)
    scope_fdbk.setTimePerDiv(0.0005)
    scope_fdbk.setChanVoltsPerDiv(1, 0.2)
    scope_fdbk.setChanVoltsPerDiv(3, 0.02)
    ai["scope.monitor"] = scope_mon
    scope_mon.setRoll(False)
    scope_mon.setRun(True)
    scope_mon.setTimePerDiv(0.0005)
    scope_mon.setChanVoltsPerDiv(1, 0.10)
    scope_mon.setChanVoltsPerDiv(2, 0.002)
    scope_mon.setChanVoltageOffset(2, -0.0726)
    apparatus.state = None
    apparatus.runStatus = False

    error_signal_response_to_eit_detuning = BasicInstrument(
        device_nickname=".".join(
            [app_nickname, "error_signal_response_to_eit_detuning"]
        ),
        tsdb_ingester=tsdb_ingester,
    )
    error_signal_response_to_eit_detuning.deviceProperties.update(
        {"conversion_factor", "unit"}
    )
    error_signal_response_to_eit_detuning.conversion_factor = None
    error_signal_response_to_eit_detuning.unit = "V_per_Hz"
    ai["error_signal_response_to_eit_detuning"] = error_signal_response_to_eit_detuning

    B_control_voltage_to_eit_detuning = BasicInstrument(
        device_nickname=".".join([app_nickname, "B_control_voltage_to_eit_detuning"]),
        tsdb_ingester=tsdb_ingester,
    )
    B_control_voltage_to_eit_detuning.deviceProperties.update(
        {"conversion_factor", "unit"}
    )
    B_control_voltage_to_eit_detuning.conversion_factor = None
    B_control_voltage_to_eit_detuning.unit = "Hz_per_V"
    ai["B_control_voltage_to_eit_detuning"] = B_control_voltage_to_eit_detuning

    # SweepSpan = 10000
    ai["rfgen"].stopFrequencySweep()
    # ai['rfgen'].setFreqFixed(freqDeltaMm2)
    # ai['rfgen'].setFreqFixed(freqZero)
    apparatus.setRFreq(freqDeltaMp2)
    # apparatus.setRFreq(freqZero)

    rfPstart = -10
    # rfPstop = 10
    # rfPowerList = np.linspace(rfPstart, rfPstop, 11)
    rfPower0 = rfPstart
    rfPower0 = 9
    apparatus.setRFAmplitude(rfPower0)

    d = getConfig(apparatus)
    instruments_config = gui.QOLDictionary(
        a=wp, name="Instruments configs", container=d
    )
    extra_tags = {}

    await initial_lock_to_eit(apparatus, extra_tags=extra_tags)

    await calibratingLockin(apparatus, extra_tags=extra_tags)
    instruments_config.container = getConfig(apparatus)
    instruments_config.display_container_dictionary()

    await responseToChangeOfBfieldControlVoltage(apparatus, extra_tags=extra_tags)
    instruments_config.container = getConfig(apparatus)
    instruments_config.display_container_dictionary()

    # await sweepRFPower(apparatus, extra_tags=extra_tags, rfPowerList=rfPowerList)

    await record_magnetometer_noise(apparatus)

    await longTermLock(apparatus, extra_tags=extra_tags)

    apparatus.state = "Done working with hardware"
    logger.info(apparatus.state)
    update_webpage(apparatus=apparatus)

    apparatus.gui_log.stop_tasks()
    task_wp_update_loop.cancel()
    logger.info("exiting main loop")
    update_webpage(apparatus=apparatus)
    return apparatus


def update_webpage(byWhom=None, apparatus=None):
    timestr = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
    clock_upd.text = f"Last update at {timestr}"
    try:
        status_line.text = apparatus.state
    except Exception:
        pass
    jp.run_task(wp.update())


async def update_webpage_loop(update_interval=1):
    while True:
        update_webpage()
        await asyncio.sleep(update_interval)


async def getPage():
    return wp


async def jp_startup():
    jp.run_task(main())


class Apparatus(BasicInstrument):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.setLockinSlope(0)

    @BasicInstrument.tsdb_append
    def getBCurrent(self):
        scope = self.instruments["scope.monitor"]
        ch = scope.getTrace(2, decimate=False)
        bCurrent = np.mean(ch.y.values)
        # mV -> mA
        return bCurrent

    def getErrAndEit(self):
        scope = self.instruments["scope.feedback"]
        # we need to freeze scope since this two values are taken on the same scope
        trg_mode = scope.getTriggerMode()
        scope.setTriggerMode("STOP")
        err = self.getErr()
        eit = self.getEIT()
        scope.setTriggerMode(trg_mode)
        return (err, eit)

    @BasicInstrument.tsdb_append
    def getErr(self):
        scope = self.instruments["scope.feedback"]
        ch1 = scope.getTrace(1, decimate=False)
        err = np.mean(ch1.y.values)
        return err

    @BasicInstrument.tsdb_append
    def getEIT(self):
        scope = self.instruments["scope.feedback"]
        ch3 = scope.getTrace(3, decimate=False)
        eit = np.mean(ch3.y.values)
        return eit

    @BasicInstrument.tsdb_append
    def setFeefback(self, val):
        freq0 = self.getRFreq()
        self.setRFreq(freq0 + val)

    @BasicInstrument.tsdb_append
    def getRFreq(self):
        rfgen = self.instruments["rfgen"]
        return rfgen.getFreqFixed()

    @BasicInstrument.tsdb_append
    def setRFreq(self, val):
        rfgen = self.instruments["rfgen"]
        rfgen.setFreqFixed(val)

    @BasicInstrument.tsdb_append
    def getRFAmplitude(self):
        rfgen = self.instruments["rfgen"]
        return rfgen.getRFAmplitude()

    @BasicInstrument.tsdb_append
    def setRFAmplitude(self, val):
        rfgen = self.instruments["rfgen"]
        return rfgen.setRFAmplitude(val)

    @BasicInstrument.tsdb_append
    def setLockinSlope(self, val):
        self.lockin_slope = val


if __name__ == "__main__":
    logger.info("========== Start up ===========")
    tsdb_client = tsdb.Client(
        "influx", "http://lumus.physics.wm.edu:8428", database="qolab"
    )
    tsdb_ingester = tsdb.Ingester(
        tsdb_client, batch=11, measurement_prefix="VAMPIRE.HighPower"
    )

    wp = jp.WebPage(delete_flag=False)
    apparatus = Apparatus(
        tsdb_ingester=tsdb_ingester,
        device_nickname="magnetometer",
    )
    apparatus.config["Device type"] = "QOL VAMPIRE HighPower magnetometer"
    apparatus.config["Device model"] = "v0.1"
    apparatus.config["FnamePrefix"] = "magnetometer"
    apparatus.config["SavePath"] = "/mnt/qol_grp_data/data.VAMPIRE.HighPower"
    # apparatus.config['SavePath'] = './data'

    d = jp.Div(
        text="Magnetometer log",
        a=wp,
        classes="text-white bg-blue-500 text-center text-xl",
    )

    div_status = jp.Div(
        classes="text-xl flex m0 p-1 space-x-4 bg-gray-300 font-mono", a=wp
    )
    clock_upd = jp.Div(
        text="Clock Loading...", classes="text-xl bg-gray-400", a=div_status
    )
    status_line = jp.Div(text="Status Loading...", classes="text-xl", a=div_status)

    # mpl.use("TkAgg")
    # apparatus = asyncio.run(main())
    jp.justpy(getPage, startup=jp_startup)



"""Basic lockin classes intended for expansion with hardware aware classes."""

from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument


class Lockin(BasicInstrument):
    """Basic lockin class with.
    Intended to be expanded in hardware specific classes.
    """

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "Lockin"
        self.config["FnamePrefix"] = "lockin"
        self.config["Device model"] = "Generic Lockin Without Hardware interface"
        self.config["FnamePrefix"] = "lockin"
        self.deviceProperties.update(
            {
                "FreqInt",
                "FreqExt",
                "Harm",
                "SinAmpl",
                "SinOffset",
                "RefPhase",
                "Sensitivity",
                "TimeConstan",
                "FilterSlope",
                "EquivalentNoiseBW",
            }
        )

    # Minimal set of methods to be implemented.
    pass


class LockinSCPI(SCPIinstr, Lockin):
    """Basic lockin class with SCPI interface.
    Intended to be expanded in hardware specific classes.

    Example
    -------
    >>> rm = pyvisa.ResourceManager()
    >>> LockinSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    pass

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Lockin.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



import platform
import re
import os
from datetime import date


def filename2os_fname(fname):
    r"""Translate Windows or Linux filename to OS dependent style.
    Takes in account the notion of 'Z:' drive on different systems.

    In particular replaces Z: <==> /mnt/qol_grp_data  and \\ <==> /

    Example
    -------
    >>> filename2os_fname("Z:\\dir1\\dir2\\file")  # when run on Linux
    '/mnt/qol_grp_data/dir1/dir2/file'
    """
    if platform.system() == "Windows":
        fname = re.sub("/mnt/qol_grp_data", "Z:", fname)
    else:
        fname = re.sub("Z:", "/mnt/qol_grp_data", fname)
        fname = re.sub(r"\\", "/", fname)

    fname = os.path.normpath(fname)
    return fname


def get_runnum(data_dir):
    r"""Reads, increments data counter and saves it back in the provided `data_dir`.
    If necessary creates counter file and full path to it.

    Examples
    --------
    >>> get_runnum('Z:\\Ramsi_EIT\\data\\')
    >>> get_runnum('/mnt/qol_grp_data/data')
    """
    data_dir = filename2os_fname(data_dir)
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)
    if not os.path.isdir(data_dir):
        print(f"ERROR: cannot create directory for data: {data_dir}")
        print("Will use current dir for storage")
        data_dir = "."

    runnumpath = os.path.join(data_dir, "autofile")
    # convert to OS dependent way
    runnumpath = filename2os_fname(runnumpath)

    if not os.path.exists(runnumpath):
        os.mkdir(runnumpath)
    runnum_file = os.path.join(runnumpath, "runnum.dat")
    runnum_file = filename2os_fname(runnum_file)

    run_number = 0
    if os.path.isfile(runnum_file):
        with open(runnum_file, "r") as f:
            content = f.readlines()
            run_number = int(content[0])
            f.close()

    # Increment it and fold if needed
    run_number = run_number + 1
    # Important: we are using five digit counters to synchronize
    # with qol_get_next_data_file.m
    if run_number > 99999:
        run_number = 0

    with open(runnum_file, "w") as f:
        f.write(f"{run_number}")
        f.close()
    return run_number


def get_next_data_file(
    prefix,
    savepath,
    run_number=None,
    datestr=None,
    date_format="%Y%m%d",
    extension="dat",
):
    """Generate a filename according to a standard naming scheme
    fname = os.path.join(savepath, f'{prefix}_{datestr}_{run_number:05d}.{extension}')
    if run_number is missing, acquires it with `get_runnum( savepath )`
    """
    if run_number is None:
        run_number = get_runnum(savepath)
    today = date.today()
    if datestr is None:
        datestr = today.strftime(date_format)
    fname = os.path.join(savepath, f"{prefix}_{datestr}_{run_number:05d}.{extension}")
    return fname


def infer_compression(fname):
    """Infers compression algorithm from filename extension"""
    compression = None  # usual suspect
    b, fext = os.path.splitext(fname)
    if fext == ".gz":
        compression = "gzip"
    elif (fext == ".bz") or (fext == ".bz2"):
        compression = "bzip"
    return compression


def save_table_with_header(
    fname,
    data,
    header="",
    comment_symbol="%",
    skip_headers_if_file_exist=False,
    item_format="e",
    item_separator="\t",
    compressionmethod=None,
    compresslevel=9,
    match_filename_to_compression=True,
):
    r"""Saves output to CSV or TSV file with specially formatted header.

    The file is appended if needed.
    It is possible to compress output file.

    Parameters
    ----------
    fname : string, full path of the saved file.
    data : array type (python or numpy).
    header : list or array of header strings to put at the beginning of the record
    comment_symbol : prefix for the header lines, default is '%'.
        Note that headers are actually prefixed with <comment_symbol> and <space>.
        Historically it is chosen as '%' to make files compatible with Matlab `load`.
        '#' symbmol is also good choice.
    skip_headers_if_file_exist : True or False (default).
        When True skip addition of headers in already existing file.
        Useful when appending to file with the same headers.
    item_format : output format like in formatted strings, examples are 'e', '.15e', 'f'
    item_separator : how to separate columns, '\t' is default.
        Natural choices are either ',' (comma) or '\t' (tab).
    compressionmethod :  compression method
        - None : no compression (default)
        - gzip : gzip method of compression
        - bzip : bzip2 method of compression
    compresslevel : 0 to 9 (default)
        Compression level as it is defined for gzip in Lib/gzip.py
        - 0 : no compression at all
        - 9 : the highest compression (default)
    match_filename_to_compression: True (default) or False
        If True changes the filename suffix in accordance with compression method,
        e.g. 'data.dat' -> 'data.dat.gz' if compression is set to 'gzip',
        otherwise assumes that users know what they do.
    """
    fname = filename2os_fname(fname)
    compression_infered = infer_compression(fname)
    if (compression_infered != compressionmethod) and match_filename_to_compression:
        if compressionmethod is None:
            fname += ".dat"
        if compressionmethod == "gzip":
            fname += ".gz"
        elif compressionmethod == "bzip":
            fname += ".bz"
    file_exist_flag = os.path.exists(fname)
    item_format = str.join("", ["{", f":{item_format}", "}"])
    _open = open  # standard file handler
    if compressionmethod == "gzip":
        import gzip

        _open = lambda fname, mode: gzip.open(  # noqa: E731
            fname, mode=mode, compresslevel=compresslevel
        )
    if compressionmethod == "bzip":
        import bz2

        _open = lambda fname, mode: bz2.open(  # noqa: E731
            fname, mode=mode, compresslevel=compresslevel
        )
    with _open(fname, mode="ab") as f:
        if not (file_exist_flag and skip_headers_if_file_exist):
            for line in header:
                f.write(f"{comment_symbol} {line}\n".encode("utf-8"))
        if data is not None:
            for r in data:
                line = item_separator.join(map(item_format.format, r))
                f.write(line.encode("utf-8"))
                f.write("\n".encode("utf-8"))
        f.close()
    return fname



import logging
import pyvisa
import numpy as np
from qolab.hardware.power_supply.keysight_e3612a import KeysightE3612A
import qolab.tsdb as tsdb

# this should be done before justpy is called or log formatter does not work
logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)

tmplogger = logging.getLogger("qolab.tsdb")
tmplogger.setLevel(logging.INFO)
logger = logging.getLogger("BfieldDriver")
logger.setLevel(logging.INFO)


class BfieldDriver(KeysightE3612A):
    """need to set power supply"""

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config["Device type"] = (
            "B field coil driver based on Keysight E3612A power supply"
        )
        self.config["Device model"] = "v0.1"
        self.deviceProperties = self.deviceProperties.union(
            {"B", "Bslope_TperA", "CoilAssignment"}
        )
        """"
        Rough magnetic field calibration of the 3 axes coils in large magnetic shield
        - Ch1: 70mA -> 650 kHz shift for delta m = 2
        - Ch2: 70mA -> 700 kHz shift for delta m = 2
        - Ch2: 70mA -> 659 kHz shift for delta m = 2
        A better calibration obtained on 20220601 see file
          20220601.magnetic_field_callibration/calibration_currentToB.dat
        """
        # B response to current in a given channel
        # Olivia's Calibration with flux gate magnetometer used from 2023/04/28
        #  +/-     0.0000007
        self._Bslope_TperA = {1: 0.00065859269, 2: 0.00070732580, 3: 0.00066754994}
        """
        # Irina's Calibration used prior 2023/04/27
        self._Bslope_TperA = {
                1: 0.0006571429,
                2: 0.0007085714,
                3: 0.0006675714
                }
        # Eugeniy's Calibration
        self._Bslope_TperA = {
                1: 0.0006574710928926532,
                2: 0.0007064314754023079,
                3: 0.0006635058865577695
                }
        """
        # assuming that Ch1 controls Bz, Ch2 -> Bx, Ch3 -> By
        self._coil_assignment = {"chX": 2, "chY": 3, "chZ": 1}

    def getBslope_TperA(self):
        return self._Bslope_TperA

    def getCoilAssignment(self):
        return self._coil_assignment

    def getB(self):
        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment["chX"]
        chY = coil_assignment["chY"]
        chZ = coil_assignment["chZ"]

        Ix = self.getChanIout(chX)
        Iy = self.getChanIout(chY)
        Iz = self.getChanIout(chZ)
        Bx = Ix * Bslope_TperA[chX]
        By = Iy * Bslope_TperA[chY]
        Bz = Iz * Bslope_TperA[chZ]
        Bmag = float(np.sqrt(Bx * Bx + By * By + Bz * Bz))
        theta = float(np.arccos(Bz / Bmag))
        phi = float(np.arctan2(By, Bx))
        theta_degree = float(theta / np.pi * 180)
        phi_degree = float(phi / np.pi * 180)
        return {
            "Bmag": Bmag,
            "theta": theta,
            "phi": phi,
            "theta_degree": theta_degree,
            "phi_degree": phi_degree,
            "Bx": float(Bx),
            "By": float(By),
            "Bz": float(Bz),
        }

    def setB(self, Bmag=50e-6, theta=0, phi=0):
        """Sets B field currents based on B (in T) and angles theta, and phi"""
        self._Bmag = Bmag
        self._theta = theta
        self._phi = phi

        Bx = Bmag * np.sin(theta) * np.cos(phi)
        By = Bmag * np.sin(theta) * np.sin(phi)
        Bz = Bmag * np.cos(theta)

        Bslope_TperA = self.getBslope_TperA()
        coil_assignment = self.getCoilAssignment()
        chX = coil_assignment["chX"]
        chY = coil_assignment["chY"]
        chZ = coil_assignment["chZ"]

        Ix = Bx / Bslope_TperA[chX]
        Iy = By / Bslope_TperA[chY]
        Iz = Bz / Bslope_TperA[chZ]

        logger.info(f"Setting {Bmag=}, {theta=} {phi=} in radians")
        logger.info(f"Setting {chX=} to {Ix}")
        logger.info(f"Setting {chY=} to {Iy}")
        logger.info(f"Setting {chZ=} to {Iz}")
        self.setChanIout(chX, Ix)
        self.setChanIout(chY, Iy)
        self.setChanIout(chZ, Iz)
        return Ix, Iy, Iz

    def setBinDegrees(self, Bmag=50e-6, theta=0, phi=0):
        logger.info(f"Setting {Bmag=},  {theta=} {phi=} in degrees")
        return self.setB(Bmag=Bmag, theta=theta / 180 * np.pi, phi=phi / 180 * np.pi)


# TSDB logger setting
tsdb_client = tsdb.Client("influx", "http://qo.physics.wm.edu:8428", database="qolab")
tsdb_ingester = tsdb.Ingester(tsdb_client, batch=11, measurement_prefix="VAMPIRE.VCSEL")

logger.info("Accessing hardware")
rm = pyvisa.ResourceManager()
instr = rm.open_resource("USB0::10893::4354::MY61001869::0::INSTR")
app_nickname = "BfieldDriver"
Bfield = BfieldDriver(
    instr,
    device_nickname=".".join([app_nickname, "b_field_driver"]),
    tsdb_ingester=None,
)


print("Use me as : Bfield.setBinDegrees(theta=Angle, phi=Angle)")



"""
Created by Eugeniy E. Mikhailov 2021/11/29
"""

from qolab.hardware.basic import BasicInstrument
from ._basic import ScopeSCPI, calcSparsingAndNumPoints
from qolab.hardware.scpi import response2numStr
from qolab.data.trace import Trace
import numpy as np
import scipy.signal
from pyvisa.constants import InterfaceType


class SDS1104X(ScopeSCPI):
    """Siglent SDS1104x scope"""

    # SDS1104x has actually 8 divisions but its behave like it has 10,
    # the grabbed trace has more points outside what is visible on the screen
    vertDivOnScreen = 10
    horizDivOnScreen = 14

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "SDS1104X"
        self.resource.read_termination = "\n"
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1000
        # desired number of points per channel, can return twice more

    def mean(self, chNum):
        # get mean on a specific channel calculated by scope
        # PAVA stands for PArameter VAlue
        qstr = f"C{chNum}:PAVA? MEAN"
        rstr = self.query(qstr)
        # reply is in the form 'C1:PAVA MEAN,3.00E-02V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=",", unit="V")
        return float(numberString)

    def getAvailableNumberOfPoints(self, chNum):
        if chNum != 1 and chNum != 3:
            # for whatever reason 'SAMPLE_NUM' fails for channel 2 and 4
            chNum = 1
        qstr = f"SAMPLE_NUM?  C{chNum}"
        rstr = self.query(qstr)
        # reply is in the form 'SANU 7.00E+01pts'
        prefix, numberString, unit = response2numStr(
            rstr, firstSeparator=" ", unit="pts"
        )
        return int(float(numberString))

    @BasicInstrument.tsdb_append
    def getSampleRate(self):
        rstr = self.query("SAMPLE_RATE?")
        # expected reply is like 'SARA 1.00E+09Sa/s'
        prefix, numberString, unit = response2numStr(
            rstr, firstSeparator=" ", unit="Sa/s"
        )
        return int(float(numberString))

    def setSampleRate(self, val):
        print("Cannot set SampleRate directly for SDS1104X")
        # it is not possible to do with this model directly
        pass

    def getRawWaveform(
        self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True
    ):
        """
        Get raw channel waveform in binary format.

        Parameters
        ----------
        chNum : int
            Scope channel to use: 1, 2, 3, or 4
        availableNpnts : int or None (default)
            Available number of points. Do not set it if you want it auto detected.
        maxRequiredPoints : int
            Maximum number of required points, if we ask less than available
            we well get sparse set which proportionally fills all available time range.
        decimate : False or True (default)
            Decimate should be read as apply the low pass filter or not, technically
            for both setting we get decimation (i.e. smaller than available
            at the scope number of points). The name came from
            ``scipy.signal.decimate`` filtering function.
            If ``decimate=True`` is used, we get all available points
            and then low-pass filter them to get ``maxRequiredPoints``
            The result is less noisy then, but transfer time from the instrument
            is longer.
            If ``decimate=False``, then it we are skipping points to get needed number
            but we might see aliasing, if there is a high frequency noise
            and sparing > 1. Unless you know what you doing, it is recommended
            to use ``decimate=True``.
        """

        rawChanCfg = {}
        if availableNpnts is None:
            # using channel 1 to get availableNpnts
            availableNpnts = self.getAvailableNumberOfPoints(1)
            rawChanCfg["availableNpnts"] = availableNpnts

        if maxRequiredPoints is None:
            maxRequiredPoints = self.maxRequiredPoints
        (
            sparsing,
            Npnts,
            availableNpnts,
            maxRequiredPoints,
        ) = calcSparsingAndNumPoints(availableNpnts, maxRequiredPoints)
        rawChanCfg["Npnts"] = Npnts
        rawChanCfg["sparsing"] = sparsing
        if decimate:
            Npnts = availableNpnts  # get all of them and decimate later
        if (sparsing == 1 and Npnts == availableNpnts) or decimate:
            # We are getting all points of the trace
            # Apparently sparsing has no effect with this command
            # and effectively uses SP=1 for any sparsing
            # but I want to make sure and force it
            cstr = "WAVEFORM_SETUP NP,0,FP,0,SP,1"
            # technically when we know Npnts and sparsing
            # we can use command from the follow up 'else' clause
        else:
            # we just ask every point with 'sparsing' interval
            # fast to grab but we could do better with more advance decimate
            # method, which allow better precision for the price
            # of longer acquisition time
            cstr = f"WAVEFORM_SETUP SP,{sparsing},NP,{Npnts},FP,0"
            # Note: it is not enough to provide sparsing (SP),
            # number of points (NP) needed to be calculated properly too!
            # From the manual
            # WAVEFORM_SETUP SP,<sparsing>,NP,<number>,FP,<point>
            # SP Sparse point. It defines the interval between data points.
            # For example:
            #   SP = 0 sends all data points.
            #   SP = 1 sends all data points.
            #   SP = 4 sends every 4th data point
            # NP — Number of points. It indicates how many points should be transmitted.
            # For example:
            #   NP = 0 sends all data points.
            #   NP = 50 sends a maximum of 50 data points.
            # FP — First point. It specifies the address of the first data point
            # to be sent.
            # For example:
            #   FP = 0 corresponds to the first data point.
            #   FP = 1 corresponds to the second data point
        self.write(cstr)

        trRaw = Trace(f"Ch{chNum}")

        qstr = f"C{chNum}:WAVEFORM? DAT2"
        # expected full reply: 'C1:WF DAT2,#9000000140.........'
        try:
            wfRaw = self.query_binary_values(
                qstr,
                datatype="b",
                header_fmt="ieee",
                container=np.array,
                chunk_size=(Npnts + 100),
            )
            if self.resource.interface_type == InterfaceType.usb:
                # Somehow on windows (at least with USB interface)
                # there is a lingering empty string which we need to flush out
                r = self.read()
                if r != "":
                    print(f"WARNING: We expected an empty string but got {r=}")
            trRaw.values = wfRaw.reshape(wfRaw.size, 1)
            if decimate and sparsing != 1:
                numtaps = 3
                # not sure it is the best case
                trRaw.values = scipy.signal.decimate(
                    trRaw.values, sparsing, numtaps, axis=0
                )
        except ValueError as err:
            # most likely we get crazy number of points
            # self.read() # flushing the bogus output of previous command
            print(f"Error {err=}: getting waveform failed for {qstr=}")
            wfRaw = np.array([])
        trRaw.config["unit"] = "Count"
        trRaw.config["tags"]["Decimate"] = decimate
        trRaw.config["tags"]["rawChanConfig"] = rawChanCfg
        return trRaw

    def getChanVoltsPerDiv(self, chNum):
        qstr = f"C{chNum}:VDIV?"
        rstr = self.query(qstr)
        # expected reply to query: 'C1:VDIV 1.04E+00V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=" ", unit="V")
        return float(numberString)

    def setChanVoltsPerDiv(self, chNum, vPerDiv):
        cstr = f"C{chNum}:VDIV {vPerDiv}"
        self.write(cstr)
        # if out of range, the VAB bit (bit 2) in the STB register to be set

    def getChanVoltageOffset(self, chNum):
        qstr = f"C{chNum}:OFST?"
        rstr = self.query(qstr)
        # expected reply to query: 'C1:OFST -1.27E+00V'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=" ", unit="V")
        return float(numberString)

    def setChanVoltageOffset(self, chNum, val):
        cstr = f"C{chNum}:OFST {val}"
        self.write(cstr)

    def getLED(self):
        """Returns binary mask of available LEDs"""
        qstr = "LED?"
        rstr = self.query(qstr)
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=" ", unit="")
        return int(numberString, 16)  # convert from hex string to integer

    def toggleRun(self):
        # SY_FP is undocumented, reverse engineered from the web interface
        self.write("SY_FP 12,1")

    @BasicInstrument.tsdb_append
    def getRun(self):
        ledStatus = self.getLED()
        return bool(ledStatus & (1 << 17))

    @BasicInstrument.tsdb_append
    def setRun(self, val):
        state = self.getRun()
        if state != val:
            self.toggleRun()

    @BasicInstrument.tsdb_append
    def getRoll(self):
        ledStatus = self.getLED()
        return bool(ledStatus & (1 << 10))

    def toggleRoll(self):
        # SY_FP is undocumented, reverse engineered from the web interface
        self.write("SY_FP 49,1")

    @BasicInstrument.tsdb_append
    def setRoll(self, val):
        rollState = self.getRoll()
        if rollState != val:
            self.toggleRoll()

    @BasicInstrument.tsdb_append
    def getTimePerDiv(self):
        qstr = "TDIV?"
        rstr = self.query(qstr)
        # expected reply to query: 'TDIV 2.00E-08S'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=" ", unit="S")
        return float(numberString)

    @BasicInstrument.tsdb_append
    def setTimePerDiv(self, timePerDiv):
        cstr = f"TDIV {timePerDiv}"
        self.write(cstr)
        # if out of range, the VAB bit (bit 2) in the STB register to be set

    @BasicInstrument.tsdb_append
    def getTrigDelay(self):
        qstr = "TRIG_DELAY?"
        rstr = self.query(qstr)
        # expected reply to query: 'TRDL -0.00E+00S'
        prefix, numberString, unit = response2numStr(rstr, firstSeparator=" ", unit="S")
        return float(numberString)

    @BasicInstrument.tsdb_append
    def setTrigDelay(self, value):
        cstr = f"TRIG_DELAY {value}"
        self.write(cstr)

    def getWaveform(
        self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True
    ):
        """
        For decimate use see ``getRawWaveform``.

        In short decimate=True is slower but more precise.
        """
        trRaw = self.getRawWaveform(
            chNum,
            availableNpnts=availableNpnts,
            maxRequiredPoints=maxRequiredPoints,
            decimate=decimate,
        )
        VoltageOffset = self.getChanVoltageOffset(chNum)
        VoltsPerDiv = self.getChanVoltsPerDiv(chNum)
        tr = trRaw
        tr.values = (
            trRaw.values * VoltsPerDiv * self.vertDivOnScreen / 250 - VoltageOffset
        )
        tr.config["unit"] = "Volt"
        tr.config["tags"]["VoltageOffset"] = VoltageOffset
        tr.config["tags"]["VoltsPerDiv"] = VoltsPerDiv
        return tr

    def getTimeTrace(self, rawChanCfg):
        availableNpnts = rawChanCfg["availableNpnts"]
        sparsing = rawChanCfg["sparsing"]
        Npnts = rawChanCfg["Npnts"]
        sampleRate = self.getSampleRate()
        timePerDiv = self.getTimePerDiv()
        trigDelay = self.getTrigDelay()
        if Npnts is None and sparsing is None:
            # using channel 1 as reference
            Npnts = self.getAvailableNumberOfPoints(1)
        tval = np.arange(Npnts) / sampleRate * sparsing
        tval = tval - timePerDiv * self.horizDivOnScreen / 2 - trigDelay
        t = Trace("Time")
        t.values = tval.reshape(tval.size, 1)
        t.config["unit"] = "s"
        t.config["tags"]["TimePerDiv"] = timePerDiv
        t.config["tags"]["TrigDelay"] = trigDelay
        t.config["tags"]["SampleRate"] = sampleRate
        t.config["tags"]["AvailableNPnts"] = availableNpnts
        t.config["tags"]["Npnts"] = availableNpnts
        t.config["tags"]["Sparsing"] = sparsing
        return t

    def getTriggerMode(self):
        # we expect NORM, AUTO, SINGLE, STOP
        res = self.query("TRIG_MODE?")
        # res is in the form 'TRMD AUTO'
        return res[5:]

    def setTriggerMode(self, val):
        # we expect NORM, AUTO, SINGLE, STOP
        self.write(f"TRIG_MODE {val}")


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("TCPIP::192.168.0.62::INSTR")
    scope = SDS1104X(instr)
    print(f"ID: {scope.idn}")
    # print(f'Ch1 mean: {scope.mean(1)}')
    print(f"Ch1 available points: {scope.getAvailableNumberOfPoints(1)}")
    print(f"Sample Rate: {scope.getSampleRate()}")
    print(f"Time per Div: {scope.getTimePerDiv()}")
    print(f"Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}")
    print(f"Ch1 Voltage Offset: {scope.getChanVoltageOffset(1)}")
    print("------ Header start -------------")
    print(str.join("\n", scope.getHeader()))
    print("------ Header ends  -------------")
    ch1 = scope.getTrace(1)
    traces = scope.getAllTraces()



import logging
import universal_tsdb as utsdb
from universal_tsdb import Client, MaxErrorsException
import functools
import time

__all__ = ["Client", "Ingester", "MaxErrorsException"]

logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)
logger = logging.getLogger("qolab.tsdb")
logger.setLevel(logging.INFO)


class Ingester(utsdb.Ingester):
    """Same as universal_tsdb.Ingester but sets measurement_prefix.

    Handy with ``@tsdb_append_metric_for_class_setter_or_getter`` decorator
    to be used in hardware controlled classes

    Parameters
    ----------
    measurement_prefix: str (default is empty)
        standard ``utsdb`` measurement becomes measurement_prefix.measurement

    Example
    -------

    >>> from qolab.hardware.basic import BasicInstrument
    >>> from qolab.tsdb import Ingester
    >>> tsdb_client = Client("influx", "http://localhost:8428", database="qolab")
    >>> tsdb_ingester = Ingester(
    >>>     tsdb_client, batch=10, measurement_prefix="experiment.title"
    >>>  )
    >>>
    >>> class InstWithLog(BasicInstrument):
    >>>     def __init__(self, *args, **kwds):
    >>>        super().__init__(*args, **kwds)
    >>>        self.config["Device type"] = "TestTSDBLogger"
    >>>        self.config["Device model"] = "v01"
    >>>        self.config["FnamePrefix"] = "test_log"
    >>>        self.config["SavePath"] = "./data"
    >>>        self.deviceProperties.update({"D"})
    >>>        self.d = 13.45
    >>>
    >>>    @BasicInstrument.tsdb_append
    >>>    def setD(self, val):
    >>>        self.d = val
    >>>
    >>>    @BasicInstrument.tsdb_append
    >>>    def getD(self):
    >>>        return self.d
    >>>
    >>> dev = InstWithLog(tsdb_ingester=tsdb_ingester, device_nickname="tester")
    >>> dev.getD()
    >>> dev.setD(3)
    >>>
    >>> tsdb_ingester.commit()
    """

    def __init__(self, client, batch=0, measurement_prefix=""):
        super().__init__(client, batch=batch)
        self.measurement_prefix = measurement_prefix

    def append(self, timestamp=None, tags=None, measurement=None, **kwargs):
        if self.measurement_prefix is None or not isinstance(
            self.measurement_prefix, str
        ):
            raise ValueError("Invalid measurement_prefix, it should be string")
        if measurement is None or not isinstance(measurement, str):
            raise ValueError("Invalid measurement, it should be string")
        qolab_measurement = ".".join((self.measurement_prefix, measurement))
        # space is illegal for measurements fields
        qolab_measurement = qolab_measurement.replace(" ", "-")
        logger.debug(f"{qolab_measurement=} {tags=}, {kwargs=}")
        return super().append(
            timestamp=timestamp, tags=tags, measurement=qolab_measurement, **kwargs
        )


def tsdb_append_metric_for_class_setter_or_getter(tsdb_logger=None):
    """Send results of setters and getters to TSDB.

    Intended to be used as decorator for setters and getters,
    i.e. it expect the wrapped function to be something like getXXX or setXXX.
    """

    def wrap(f):
        @functools.wraps(f)
        def wrapper(*args, **kwds):
            if f.__name__[0:3] != "get" and f.__name__[0:3] != "set":
                logger.warning(
                    f"Do not know how to work with {f.__name__}"
                    + ", it is neither setXXX nor getXXX"  # noqa: W503
                )
                ret = f(*args, **kwds)
                return ret

            cls = args[0]
            action = f.__name__[0:3]
            var_name = f.__name__[3:]
            val = None
            if cls.config["DeviceNickname"] is not None:
                device_type = cls.config["DeviceNickname"]
            else:
                device_type = cls.config["Device type"]
            if action == "get":
                """getter"""
                val = f(*args, **kwds)
                ts = time.time()
                ret = val
            else:
                """setter"""
                val = args[1]
                ts = time.time()
                ret = f(*args, **kwds)

            logger.debug(f"function {f.__name__} {action} {var_name} = {val}")
            ts_ms = int(ts * 1000)
            fields = {var_name: val}
            try:
                if cls.tsdb_ingester is not None:
                    cls.tsdb_ingester.append(
                        ts_ms,
                        measurement=device_type,
                        tags={"action": action},
                        **fields,
                    )
            except ValueError as err:
                logger.error(f"{err=} in function {f.__name__}: {var_name} = {val}")
            return ret

        return wrapper

    return wrap


if __name__ == "__main__":
    from qolab.hardware.basic import BasicInstrument

    tsdb_client = Client("influx", "http://localhost:8428", database="qolab")
    tsdb_ingester = Ingester(
        tsdb_client, batch=10, measurement_prefix="experiment.title"
    )

    class InstWithLog(BasicInstrument):
        def __init__(self, *args, **kwds):
            super().__init__(*args, **kwds)
            self.config["Device type"] = "TestTSDBLogger"
            self.config["Device model"] = "v01"
            self.config["FnamePrefix"] = "test_log"
            self.config["SavePath"] = "./data"
            self.deviceProperties.update({"D"})
            self.d = 13.45

        @BasicInstrument.tsdb_append
        def setD(self, val):
            self.d = val

        @BasicInstrument.tsdb_append
        def getD(self):
            """get D variable"""
            return self.d

    dev = InstWithLog(tsdb_ingester=tsdb_ingester, device_nickname="tester")
    dev.getD()
    dev.setD(3)

    tsdb_ingester.commit()



import pyvisa
from qolab.hardware.scope import SDS1104X

if __name__ == "__main__":
    print("Testing SDS1104X")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("TCPIP::192.168.0.61::INSTR")
    scope = SDS1104X(instr)
    print(f"ID: {scope.idn}")
    print(f"Ch1 mean: {scope.mean(1)}")
    print(f"Ch1 available points: {scope.getAvailableNumberOfPoints(1)}")
    print(f"Sample Rate: {scope.getSampleRate()}")
    print(f"Time per Div: {scope.getTimePerDiv()}")
    print(f"Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}")
    print(f"Ch1 Voltage Offset: {scope.getChanOffset(1)}")
    ch1 = scope.getTrace(1)
    ch1.plot()



"""
provide basic class to operate SCPI capable instruments
"""

import re
import logging
import time

logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)
logger = logging.getLogger("qolab.hardware.scpi")
logger.setLevel(logging.INFO)


def response2numStr(strIn, firstSeparator=None, unit=None):
    """Parses non standard SCPI reply.

    Often an instrument reply is in the form 'TDIV 2.00E-08S'
    (for example Siglent Scope).
    I.e. "<prefix><firstSeparator><numberString><unit> where
    prefix='TDIV', firstSeparator=' ', numberString='2.00E-08', unit='S'

    Parameters
    ----------
    prefix : str
        reply prefix, e.g. 'TDIV'
    firstSeparator : str
        separator between numerical part and prefix, e.g. ' '
    unit: str
        unit used in the reply, e.g. 'S'

    Returns
    -------
    (prefix, numberString, unit) tuple
    """
    prefix = None
    rstr = strIn
    if firstSeparator is not None and firstSeparator != "":
        spltStr = re.split(firstSeparator, strIn)
        prefix = spltStr[0]
        rstr = spltStr[1]
    if unit is not None and unit != "":
        spltStr = re.split(unit, rstr)
        if len(spltStr) == 1:
            unit = None
        numberString = spltStr[0]
    else:
        numberString = rstr
    return (prefix, numberString, unit)


class SCPI_PROPERTY(property):
    """Overrides 'property' class and makes it suitable for SCPI set and query notation.

    Adds ability to log into TSDB.
    Works within SCPIinstr class since it assumes that owner has query() and write().

    Parameters
    ----------
    scpi_prfx : str or None (default)
        SCPI command prefix to get/set property, for example 'FreqInt'
        is internally transformed to
        query 'FreqInt?' and setter 'FreqInt {val}'.
        It could be set as the explicit query and set format string list:
        ['AUXV? 1', 'AUXV 1,{}'] where {} is place holder for set value
    ptype :  type, str is default
        property type 'str', 'int', 'float', ...
    doc : str or None (default)
        short description of property, for example 'Internal lockin frequency'
    no_getter: True of False (default)
        does property has a getter, some properties has no getter,
        e.g. output voltage of a DAC
    no_setter : True of False (default)
        does property has a setter, some properties has no setter,
        e.g. measurement of external voltages for an ADC
    tsdb_logging : True (default) or False
        do we log get/set commands result/argument to TSDB

    Examples
    --------
    x = SCPI_PROPERTY(scpi_prfx='SETX', ptype=str, doc='property X', tsdb_logging=False)

    """

    def __init__(
        self,
        scpi_prfx=None,
        ptype=str,
        doc=None,
        no_getter=False,
        no_setter=False,
        tsdb_logging=True,
    ):
        self.no_getter = no_getter
        self.no_setter = no_setter
        self.tsdb_logging = tsdb_logging
        if no_getter:
            fget = None
        else:
            fget = self.get_scpi
        if no_setter:
            fset = None
        else:
            fset = self.set_scpi
        super().__init__(fget=fget, fset=fset)
        self.scpi_prfx = scpi_prfx
        self.ptype = ptype
        self.__doc__ = doc
        if isinstance(scpi_prfx, str):
            self.scpi_prfx_get = "".join([self.scpi_prfx, "?"])
            self.scpi_prfx_set = "".join([self.scpi_prfx, " {}"])
        elif isinstance(scpi_prfx, list):
            if len(scpi_prfx) != 2:
                raise ValueError(
                    f"{scpi_prfx=}, should be list with exactly two elements"
                )
            self.scpi_prfx_get = self.scpi_prfx[0]
            self.scpi_prfx_set = self.scpi_prfx[1]
        else:
            raise ValueError(f"{scpi_prfx=}, it should be either str or list type")

        if not isinstance(self.scpi_prfx_get, str):
            raise ValueError(f"{self.scpi_prfx_get=}, it should be str type")
        if not isinstance(self.scpi_prfx_set, str):
            raise ValueError(f"{self.scpi_prfx_set=}, it should be str type")

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = "_" + name

    def log_to_tsdb(self, owner, action=None, val=None):
        if owner.tsdb_ingester is None or not self.tsdb_logging:
            return
        if owner.config["DeviceNickname"] is not None:
            measurement = owner.config["DeviceNickname"]
        else:
            measurement = owner.config["Device type"]
        ts = time.time()
        ts_ms = int(ts * 1000)
        var_name = self.public_name
        tags = {"action": action}
        fields = {var_name: val}
        try:
            msg = f"{ts_ms=}, {measurement=}, {tags=}, {fields=}"
            logger.debug(msg)
            owner.tsdb_ingester.append(
                ts_ms, measurement=measurement, tags=tags, **fields
            )
        except ValueError as err:
            logger.error(f"{err=}: cannot log to TSDB {var_name} = {val}")

    def get_scpi(self, owner):
        val = self.ptype(owner.query(f"{self.scpi_prfx_get}"))
        self.log_to_tsdb(owner, action="get", val=val)
        return val

    def set_scpi(self, owner, val):
        cstr = self.scpi_prfx_set.format(val)
        owner.write(cstr)
        self.log_to_tsdb(owner, action="set", val=val)

    def __repr__(self):
        sargs = []
        sargs.append(f"scpi_prfx={self.scpi_prfx}")
        sargs.append(f"ptype={self.ptype}")
        sargs.append(f"doc={self.__doc__}")
        sargs.append(f"no_getter={self.no_getter}")
        sargs.append(f"no_setter={self.no_setter}")
        sargs.append(f"tsdb_logging={self.tsdb_logging}")
        sargs = ", ".join(sargs)
        s = "".join([f"{self.__class__.__name__}(", sargs, ")"])
        return s


class SCPIinstr:
    """Basic class which support SCPI commands.

    Intended to be a parent for more hardware aware classes.

    Example
    -------
    >>> rm = pyvisa.ResourceManager()
    >>> SCPIinstr(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    def __init__(self, resource):
        self.resource = resource

        # convenience pyvisa functions
        self.read_bytes = self.resource.read_bytes
        self.read_binary_values = self.resource.read_binary_values
        self.query_binary_values = self.resource.query_binary_values

    def write(self, *args):
        logger.debug(f"write with args: {args}")
        return self.resource.write(*args)

    def read(self, *args):
        logger.debug(f"read with args: {args}")
        return self.resource.read(*args)

    def query(self, *args, **kwds):
        logger.debug(f"query with args: {args}")
        return self.resource.query(*args, **kwds)

    @property
    def idn(self):
        return self.query("*IDN?")

    def clear_status(self):
        self.write("*CLS")

    def set_event_status_enable(self):
        self.write("*ESE")

    def query_event_status_enable(self):
        return self.query("*ESE?")

    def query_event_status_register(self):
        return self.query("*ESR?")

    def set_wait_until_finished(self):
        self.write("*OPC")

    def wait_until_finished(self):
        return self.query("*OPC?")

    def reset(self):
        self.write("*RST")

    def set_service_request_enable(self):
        self.write("*SRE")

    def query_service_request_enable(self):
        return self.query("*SRE?")

    def query_status_byte(self):
        return self.query("*STB?")

    def self_test_result(self):
        return self.query("*TST?")

    def wait(self):
        self.write("*WAI")


if __name__ == "__main__":
    from qolab.hardware.basic import BasicInstrument

    class DummyInstrument(BasicInstrument):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        # in order to check SCPI_PROPERTY we need to implement write and query
        def write(self, str):
            print(f"write: {str=}")

        def query(self, str):
            print(f"query: {str=}")
            return "123"

        x = SCPI_PROPERTY(
            scpi_prfx="SETX", ptype=str, doc="property X", tsdb_logging=False
        )
        y = SCPI_PROPERTY(scpi_prfx="SETY", ptype=int, no_setter=True, doc="property Y")
        z = SCPI_PROPERTY(scpi_prfx="SETY", ptype=int, no_getter=True, doc="property Z")

    c1 = DummyInstrument()
    c1.deviceProperties.update({"x", "y"})
    c1.getConfig()
    c2 = DummyInstrument()



from qolab.hardware.basic import BasicInstrument
from ._basic import PowerSupplySCPI
import time


class KeysightE3612A(PowerSupplySCPI):
    """Keysight E3612A power supply"""

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.resource.read_termination = "\n"
        self.config["Device model"] = "Keysight E3612A"
        self.numberOfChannels = 3
        self.deviceProperties.update({"OpMode"})
        self.channelProperties = {
            "IsOn",
            "Regulation",
            "Vout",
            "Vlimit",
            "Iout",
            "Ilimit",
            "dV",
            "dI",
        }
        self.deffaultChannelR = 47
        # used if no empirical way to calculate it via Vout/Iout

    def getChandV(self, chNum):
        """Get voltage precision per channel.

        Obtained from data sheet.
        Alternative estimate is by assuming 14 bit precision from the maximum reading.
        """
        if chNum == 1:
            return 0.24e-3
        return 1.5e-3

    def getChandI(self, chNum):
        """Current precision per channel.

        Obtained from data sheet.
        Alternative estimate is by assuming 14 bit precision from the maximum reading.
        """
        if chNum == 1:
            return 0.2e-3
        return 0.160e-3  # see specification for high current > 20mA

    def getOpMode(self):
        """Queries power supply operation mode, returns OFF|PAR|SER|TRAC.

        OFF stands for independent channels
        """
        qstr = "OUTP:PAIR?"
        rstr = self.query(qstr)
        return rstr

    def setOpMode(self, val):
        """Sets power supply operation mode, returns OFF|PAR|SER|TRAC.

        OFF stands for independent channels.
        """
        cmnd = f"OUTP:PAIR {val}"
        self.write(cmnd)

    def setChanOn(self, chNum):
        """Power up channel output"""
        self.write(f"OUTP ON,(@{chNum})")

    def setChanOff(self, chNum):
        """Power down channel output"""
        self.write(f"OUTP OFF,(@{chNum})")

    @BasicInstrument.tsdb_append
    def getChanIsOn(self, chNum):
        """Queries channel output state"""
        qstr = f"OUTP? (@{chNum})"
        rstr = self.query(qstr)
        return bool(float(rstr))

    @BasicInstrument.tsdb_append
    def getChanRegulation(self, chNum):
        """Queries channel output regulation

        0 - The output is off and unregulated
        1 - The output is CC (constant current) operating mode
        2 - The output is CV (constant voltage) operating mode
        3 - The output has hardware failure
        """
        qstr = f"STAT:QUES:INST:ISUM{chNum}:COND?"
        rstr = self.query(qstr)
        return int(rstr)

    @BasicInstrument.tsdb_append
    def getChanVout(self, chNum):
        qstr = f"MEAS:VOLT? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def getChanVlimit(self, chNum):
        qstr = f"SOUR:VOLT? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setChanVlimit(self, chNum, val):
        if val < 0:
            val = 0
        if chNum == 1 and val > 6.180:
            val = 6.180
        if (chNum == 2 or chNum == 3) and val > 25.750:
            val = 25.750
        cmnd = f"SOURCe:VOLT {val},(@{chNum})"
        self.write(cmnd)

    @BasicInstrument.tsdb_append
    def getChanIout(self, chNum):
        qstr = f"MEAS:CURR? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    def setChanIout_mA(self, chNum, val, **kwds):
        """Set current in mA.

        Calls setChanIout with val converted from mA to A.
        """
        return self.setChanIout(chNum, val / 1000.0, **kwds)

    @BasicInstrument.tsdb_append
    def setChanIout(self, chNum, val, currentHeadRoom=1e-3, dwellTime=0.3):
        """Set current in the channel.

        We will tune Vout to achieve desired Iout.
        Generally setting current limit will maintain current near
        but not exact to desired value.
        Since Vlimit can be set with good precision,
        this function will try tune Vlimit until the Idesired is reached.
        """
        iDesired = val
        # self.setChanIlimit(chNum, val+currentHeadRoom)
        # Here, we assume that hook up is already made,
        # so we can estimate source resistance.
        # So the protocol is the following:
        #  find R -> calculate required Vout for the desired Idesired
        #  -> set Vlimit to reach desired Vout and Iout
        # In general, once we estimate resistance of the load + source,
        # we do not need to anything extra.
        # But there is a problem: for a given Vlimit setting, the actual Vout
        # is slightly off.
        # We will assume that Vlimit = R*Iout + Vo = Vout + Vo,
        # i.e. linear approximation
        for i in range(10):
            iOut = self.getChanIout(chNum)
            if abs(iOut - iDesired) <= self.getChandI(chNum):
                break
            vOut = self.getChanVout(chNum)
            if self.getChanRegulation(chNum) == 2:  # i.e. CV mode
                vLimit = self.getChanVlimit(chNum)
                Vo = vLimit - vOut
            else:
                Vo = 0
            if (iOut == 0) or (
                vOut <= 0.001
            ):  # when vOut set to 0 the numbers are misreported
                R = self.deffaultChannelR  # some default
            else:
                R = vOut / iOut
            vDesired = R * iDesired
            self.setChanVlimit(chNum, vDesired + Vo)
            time.sleep(dwellTime)

    @BasicInstrument.tsdb_append
    def getChanIlimit(self, chNum):
        qstr = f"SOURce:CURR? (@{chNum})"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setChanIlimit(self, chNum, val):
        """Set current limit, seems to be >=0.002 for Ch1 and >=0.001 for Ch2 and Ch3"""
        if chNum == 1 and val < 0.002:
            val = 0.002
        if chNum == 1 and val > 5.150:
            val = 5.150
        if (chNum == 2 or chNum == 3) and val < 0.001:
            val = 0.001
        if (chNum == 2 or chNum == 3) and val > 1.030:
            val = 1.030
        cmnd = f"SOURCe:CURR {val},(@{chNum})"
        self.write(cmnd)


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("USB0::10893::4354::MY61001869::0::INSTR")
    ps = KeysightE3612A(instr)
    print("------ Header start -------------")
    print(str.join("\n", ps.getHeader()))
    print("------ Header ends  -------------")



from ._basic import LockinSCPI
from qolab.hardware.scpi import SCPI_PROPERTY


class SRS_SR865A(LockinSCPI):
    """SRS SR865A lockin"""

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "SRS SR865A"
        self.resource.read_termination = "\n"
        self.deviceProperties.update(
            {"TimeBaseMode", "AuxOut1", "AuxOut2", "AuxOut3", "AuxOut4"}
        )

    FreqInt = SCPI_PROPERTY(
        scpi_prfx="FreqInt", ptype=float, doc="Internal LO frequency"
    )
    FreqExt = SCPI_PROPERTY(
        scpi_prfx="FreqExt", ptype=float, doc="External LO frequency", no_setter=True
    )
    Harm = SCPI_PROPERTY(scpi_prfx="Harm", ptype=float, doc="Harmonic of the LO")
    SinAmpl = SCPI_PROPERTY(scpi_prfx="SLVL", ptype=float, doc="Modulation amplitude")
    SinOffset = SCPI_PROPERTY(scpi_prfx="SOFF", ptype=float, doc="Modulation offset")
    EquivalentNoiseBW = SCPI_PROPERTY(
        scpi_prfx="ENBW",
        ptype=float,
        doc="Equivalent Noise BandWidth (it is not 3dB BW, see the manual)",
        no_setter=True,
    )
    TimeBaseMode = SCPI_PROPERTY(
        scpi_prfx="TBMODE",
        ptype=int,
        doc="10 MHz reference mode: 0 - Auto, 1 - Internal",
    )
    RefPhase = SCPI_PROPERTY(scpi_prfx="PHAS", ptype=float, doc="LO/reference phase")

    Sensitivity = SCPI_PROPERTY(
        scpi_prfx="SCAL",
        ptype=int,
        doc="""
    Sensitivity:
        0:   1  V (uA),  6:  10 mV (nA),  ... 27: 1 nV (fA)
        1: 500 mV (nA),  7:   5 mV (nA),
        2: 200 mV (nA),  8:   2 mV (nA),
        3: 100 mV (nA),  9:   1 mV (nA),
        4:  50 mV (nA), 10: 500 uV (pA),
        5:  20 mV (nA), 11: 200 uV (pA),
    """,
    )
    TimeConstan = SCPI_PROPERTY(
        scpi_prfx="OFLT",
        ptype=int,
        doc="""
    Time constant:
        0:   1 uS,    6:   1 mS, ..... , 21: 30 kS
        1:   3 uS,    7:   3 mS,
        2:  10 uS,    8:  10 mS,
        3:  30 uS,    9:  30 mS,
        4: 100 uS,   10: 100 mS,
        5: 300 uS,   11: 300 mS,
    """,
    )
    FilterSlope = SCPI_PROPERTY(
        scpi_prfx="OFSL",
        ptype=int,
        doc="""
    Output Filter slope:
        0:  6dB/Oct
        1: 12dB/Oct
        2: 18dB/Oct
        3: 24dB/Oct
    """,
    )

    AuxOut1 = SCPI_PROPERTY(
        scpi_prfx=["AUXV? 0", "AUXV 0,{}"],
        ptype=float,
        doc="Voltage at Auxilarly output 1",
    )
    AuxOut2 = SCPI_PROPERTY(
        scpi_prfx=["AUXV? 1", "AUXV 1,{}"],
        ptype=float,
        doc="Voltage at Auxilarly output 2",
    )
    AuxOut3 = SCPI_PROPERTY(
        scpi_prfx=["AUXV? 2", "AUXV 2,{}"],
        ptype=float,
        doc="Voltage at Auxilarly output 3",
    )
    AuxOut4 = SCPI_PROPERTY(
        scpi_prfx=["AUXV? 3", "AUXV 3,{}"],
        ptype=float,
        doc="Voltage at Auxilarly output 4",
    )

    AuxIn1 = SCPI_PROPERTY(
        scpi_prfx=["OAUX? 0", ""],
        ptype=float,
        doc="Voltage at Auxilarly input 1",
        no_setter=True,
    )
    AuxIn2 = SCPI_PROPERTY(
        scpi_prfx=["OAUX? 1", ""],
        ptype=float,
        doc="Voltage at Auxilarly input 2",
        no_setter=True,
    )
    AuxIn3 = SCPI_PROPERTY(
        scpi_prfx=["OAUX? 2", ""],
        ptype=float,
        doc="Voltage at Auxilarly input 3",
        no_setter=True,
    )
    AuxIn4 = SCPI_PROPERTY(
        scpi_prfx=["OAUX? 3", ""],
        ptype=float,
        doc="Voltage at Auxilarly input 4",
        no_setter=True,
    )


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("TCPIP::192.168.0.51::INSTR")
    lockin = SRS_SR865A(instr)
    print("------ Header start -------------")
    print(str.join("\n", lockin.getHeader()))
    print("------ Header ends  -------------")



"""
by Rob Behary and Eugeniy Mikhailov 2024/06/17
"""

from .sds1104x import SDS1104X
from qolab.hardware.basic import BasicInstrument
from qolab.hardware.scpi import response2numStr


class SDS2304X(SDS1104X):
    """Siglent SDS2304x scope"""

    # unlike SDS1104X, number of divisions matches what is seen on the screen
    vertDivOnScreen = 8
    horizDivOnScreen = 10

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "SDS2304X"
        self.resource.read_termination = "\n"
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1000
        # desired number of points per channel, can return twice more

    @BasicInstrument.tsdb_append
    def getTimePerDiv(self):
        qstr = "TDIV?"
        rstr = self.query(qstr)
        # Siglent claims that this model should have same commands as SDS1104X
        # However response is different.
        # For example we got '2.00E-08S' instead 'TDIV 2.00E-08S'
        # expected reply to query: '2.00E-08S'
        prefix, numberString, unit = response2numStr(
            rstr, firstSeparator=None, unit="S"
        )
        return float(numberString)



"""
This unit talks to i-Series controllers sold by Omega and Newport.
They are often can be connected via iServer interface
Specifically, i853 units with LAN connection, but should be ok with many others.

Querying of this controller is slo-o-o-w at least 0.2 second and can be longer.
"""

from qolab.hardware.basic import BasicInstrument
from cachetools import cached, TTLCache
import socket
import logging

logger = logging.getLogger("qolab.hardware.i_server.i800")


class I800(BasicInstrument):
    """Newport i800 series controller, should work with similar Omega controllers

    Parameters
    ----------
    host : str
        hostname or IP of the controller unit
        the factory preset is '192.168.1.200' (default)
    port : int
        socket port (1000 i: default) which accept HTTP POST requests

    Example
    -------

    >>> tc = I800(host='192.168.1.200', port=1000)
    >>> print(tc.getConfig())
    """

    TTL_MEASURED = 30  # Time To Live for device measured things, i.e. Temperature
    TTL_SEATABLES = 600  # Time To Live for user seatables, i.e. SetPoints, Gains, etc

    def __init__(self, *args, host="192.168.1.200", port=1000, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.host = host
        self.port = port
        self.modbus_address = "01"
        self.cmd_start_marker = "*"
        self.config["Device type"] = "TemperatureController"
        self.config["Device model"] = "i800"
        self.config["FnamePrefix"] = "temperature"
        self.deviceProperties.update({"Temperature", "SetPoint1", "SetPoint2"})

    def query(self, cmnd, trials=10):
        modbus_cmnd = f"{self.modbus_address}{cmnd}"
        qstr = f"POST / HTTP/1.1\r\n\r\n{self.cmd_start_marker}{modbus_cmnd}\r\n"
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.host, self.port))
        s.send(qstr.encode("ascii"))
        reply = s.recv(100).decode("ascii")
        s.close()
        rlist = reply.split()
        # occasionally there is more than one reply, it also removes \r
        lreply = rlist[-1]
        # we will use the last one
        if lreply[0:5] != f"{modbus_cmnd}":
            # check the proper echo response
            logger.warning(
                f"Warning: expected {modbus_cmnd} but got {lreply[0:5]}"
                + f" in full set {reply}"  # noqa: W503
            )
            if trials > 0:
                return self.query(cmnd, trials - 1)
            return None
        return lreply[5:]

    @BasicInstrument.tsdb_append
    @cached(cache=TTLCache(maxsize=1, ttl=TTL_MEASURED))
    def getTemperature(self):
        command = "X01"
        # give decimal representation (X) of the temperature (01 address)
        reply = self.query(command)
        if reply is not None:
            return float(reply)
        else:
            return float("nan")

    def setPoinStr2value(self, spStr):
        raw = int(spStr, 16)
        if raw & (1 << 23):
            sign = -1
        else:
            sign = 1

        if raw & (0b1 << 20):
            scale = 1
        elif raw & (0b10 << 20):
            scale = 10
        elif raw & (0b11 << 20):
            scale = 100
        elif raw & (0b100 << 20):
            scale = 100
        else:
            logger.error(
                f"Error: unknown decimal point position in decoded {spStr} {bin(raw)}"
            )
            return float("nan")

        val = raw & 0xFFFFF
        return float(sign * val / scale)

    @BasicInstrument.tsdb_append
    @cached(cache=TTLCache(maxsize=1, ttl=TTL_SEATABLES))
    def getSetPoint1(self):
        command = "R01"
        reply = self.query(command)
        if reply is not None:
            return self.setPoinStr2value(reply)
        else:
            return float("nan")

    @BasicInstrument.tsdb_append
    @cached(cache=TTLCache(maxsize=1, ttl=TTL_SEATABLES))
    def getSetPoint2(self):
        command = "R02"
        reply = self.query(command)
        if reply is not None:
            return self.setPoinStr2value(reply)
        else:
            return float("nan")


if __name__ == "__main__":
    tc = I800()
    print(tc.getConfig())



from qolab.hardware.basic import BasicInstrument
from ._basic import Multimeter
from pyvisa import constants as pyvisa_constants
import time


class BK_5491(Multimeter):
    r"""BK 5491 multimeter

    Make sure to switch off the ECHO at the multimeter communication setup.

    Example
    -------
    >>> rm = pyvisa.ResourceManager()
    >>> instr=rm.open_resource('ASRL/dev/ttyUSB0::INSTR')
    >>> multimeter = BK_5491(instr)
    >>> print("------ Header start -------------")
    >>> print(str.join("\n", multimeter.getHeader()))
    >>> print("------ Header ends  -------------")
    """

    def __init__(self, resource, *args, **kwds):
        super().__init__(*args, **kwds)
        self.resource = resource
        self.config["Device model"] = "BK 5491"
        self.resource.read_termination = "\r\n"
        self.resource.baud_rate = 9600
        self.resource.data_bits = 8
        self.resource.parity = pyvisa_constants.Parity.none
        self.resource.stop_bits = pyvisa_constants.StopBits.one
        self.resource.timeout = 5000

        self.read = self.resource.read
        # we need to work around the prompts which BK_5491 sends during
        # communication to mimic SCPI
        # if you need raw connection use self.resource.write or self.resource.query
        # self.write = self.resource.write
        # self.query = self.resource.query
        self.read_bytes = self.resource.read_bytes
        self.read_binary_values = self.resource.read_binary_values
        self.query_binary_values = self.resource.query_binary_values

        self.switchTime = 0.5  # switch time in seconds for Function/Measurement change
        self.deviceProperties.update({"Function"})

    def isPrompt(self, string):
        if string[1] == ">":
            return True
        return False

    def isPromptGood(self, prompt):
        if prompt[0:2] == "=>":
            return True
        print(f"Error detected {prompt=}")
        return False

    def write(self, cmd_string):
        """Write/send command to instrument"""
        return self._readwrite(cmd_string, expect_reply=False, Nattemts=1)

    def query(self, cmd_string):
        """Query instrument with command"""
        return self._readwrite(cmd_string, expect_reply=True, Nattemts=5)

    def _readwrite(self, cmd_string, expect_reply=True, Nattemts=5):
        """
        Send command to instrument or query it readings (which is also a command)

        BK_5491 is not a SCPI instrument,
        so we get some replies (prompts ``*>``, ``=>``, etc)
        even if we just send a command not a query. So we have to work around this.
        """
        self.resource.read_bytes(self.resource.bytes_in_buffer)  # clear read buffer
        # print(f"dbg: {cmd_string=}")
        self.resource.write(cmd_string)
        if expect_reply:
            reply = self.resource.read()  # this should be result
            # print(f"dbg: {reply=}")
            if self.isPrompt(reply):
                prompt = reply
                if prompt[0] == "@":
                    if Nattemts >= 2:
                        """
                        print(
                        "dbg: numeric reading is not available yet"
                        + ", attempt one more time"
                        )
                        """
                        time.sleep(self.switchTime)
                        return self._readwrite(
                            cmd_string, expect_reply=expect_reply, Nattemts=Nattemts - 1
                        )
                print(
                    f"Error: we ask {cmd_string=}"
                    + f' and got prompt "{reply}" instead of result'  # noqa: W503
                )
                return None
        else:
            reply = None
        prompt = self.resource.read()  # this should be prompt
        if not self.isPromptGood(prompt):
            print(f'Error: expected good prompt but got "{prompt=}"')
        return reply

    def getReading(self):
        """Report current measurement displayed on the first/main display"""
        ret_string = self.query("R1")
        # print(f'dbg: getReading received "{ret_string}"')
        return float(ret_string)

    """
    BK_5491 has two displays which could be set and read separately,
    here we use only setting of the 1st display (prefix S1 below)
    to set the measurement and read it later.
    It is also possible to set range, but I prefer to leave in to
    the front panel user.
    If this is needed it would be followed by 3 symbols specifiers as
    outline in the manual.
    """

    @BasicInstrument.tsdb_append
    def getVdc(self):
        self.write("S10")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getVac(self):
        self.write("S11")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getAdc(self):
        self.write("S14")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getAac(self):
        self.write("S15")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getResistance(self):
        self.write("S12")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getResistance4Wires(self):
        self.write("S13")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getDiode(self):
        self.write("S16")
        return self.getReading()

    @BasicInstrument.tsdb_append
    def getFreq(self):
        self.write("S17")
        return self.getReading()

    """
    With BK_5491
    It is possible to send "key presses" like they are coming from the front panel
        K1 - Vdc
        K2 - Adc
        K3 - Vac
        K4 - Aac
        K5 - Resistance
        K6 - Diode
        K7 - Frequency (Hz)
        K8 - Auto
        K9 - Up key
        K10 - Down key
        K11 - MinMax key
        K12 - Hold key
        K13 - Local (manual does not specify, but it works this way)
        K14 - Rel key
        K15 - Shift key
        K16 - 2nd key
        K17 - Vdc and Vac keys simultaneously
        K18 - Adc and Aac keys simultaneously
        K19 - Shift then Up keys (increasing the intensity of the VFD display)
        K20 - Shift then Down keys (decreasing the intensity of the VFD display)
    """

    def toLocal(self):
        self.sendCmd("K13", expect_reply=False)

    def getFunction(self):
        reply = self.query("R0")
        """
        According to the manual:
        The reply is in 10 digits in the form <h1><h2><g1><g2><v><x><f1><r1><f2><r2>
        the <f1> and <f2> correspond to measurement/function of the 1st and 2nd display.
        Looks like it some sort of a lie, since in our BK_5491A with firmware v1.23,3
        we get back either 11 digits (if both displays are on) or 9 if only 1st display
        is on.
        We are concerned with 1st (primary) display
        """
        if len(reply) == 9:
            f1 = reply[7]
        elif len(reply) == 11:
            f1 = reply[7]
        else:
            return "Unknown"
        print(f1)
        if f1 == "0":
            return "Vdc"
        elif f1 == "1":
            return "Vac"
        elif f1 == "2":
            return "Resistance"
        elif f1 == "3":
            return "Resistance4Wires"
        elif f1 == "4":
            return "Adc"
        elif f1 == "5":
            return "Aac"
        elif f1 == "6":
            return "Diode"
        elif f1 == "7":
            return "Frequency"
        elif f1 == "8":
            return "V(ac+dc)"
        elif f1 == "9":
            return "A(ac+dc)"
        elif f1 == "A":
            return "Continuity"
        else:
            return "Unknown"


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("ASRL/dev/ttyUSB0::INSTR")
    multimeter = BK_5491(instr)
    print("------ Header start -------------")
    print(str.join("\n", multimeter.getHeader()))
    print("------ Header ends  -------------")



from qolab.file_utils import save_table_with_header, infer_compression
import datetime
import numpy as np
import yaml
import pandas
import os


def headerFromDictionary(d, prefix=""):
    """Converts dictionary to YAML format with optional prefix for every line."""
    header = []
    tail = yaml.dump(d, default_flow_style=False, sort_keys=False)
    tail = tail.split("\n")
    header.extend(tail)
    prefixed_header = [prefix + line for line in header]
    return prefixed_header


def from_timestamps_to_dates(timestamps):
    """Formats timestamps to datetime format"""
    dates = [datetime.datetime.fromtimestamp(float(ts)) for ts in timestamps]
    return dates


def loadTraceRawHeaderAndData(fname, tryCompressedIfMissing=True):
    """Load trace file and return header and data.

    Parameters
    ----------
    fname: str or Path
        file name path

    tryCompressedIfMissing: False or True (default)
        - False : use file name as is
        - True : if the provided filename is missing
            attempts to load a compressed version of filename.
            e.g. look for 'data.dat.gz', 'data.dat.bz', etc
    """
    headerstr = []
    data = None
    if (not os.path.exists(fname)) and tryCompressedIfMissing:
        # attempt to locate compressed file for the missing base file
        for ext in ["gz", "bz", "bz2"]:
            if os.path.exists(fname + "." + ext):
                fname += "." + ext
                break
    # we will try to guess if the file compressed
    _open = open
    compression = infer_compression(fname)
    if compression == "gzip":
        # TODO improve detection by reading magic bytes:
        # gzip files have first 2 bytes set to b'\x1f\x8b'
        import gzip

        _open = gzip.open
    elif compression == "bzip":
        # TODO improve detection by reading magic bytes:
        # bzip files have first 2 bytes set to b'BZ'
        import bz2

        _open = bz2.open
    with _open(fname, mode="rb") as tracefile:
        # Reading yaml header prefixed by '% '
        # It sits at the top and below is just data in TSV format
        while True:
            ln = tracefile.readline()
            if ln[0:2] == b"% ":
                headerstr.append(ln[2:].decode("utf-8"))
            else:
                break
        header = yaml.load(str.join("\n", headerstr), Loader=yaml.BaseLoader)
        # now we load the data itself
        tracefile.seek(0)  # rewind file to the beginning
        # Note: pandas reads csv faster by factor of 8 then numpy.genfromtxt
        df = pandas.read_csv(tracefile, comment="%", delimiter="\t", header=None)
        data = df.to_numpy()
    return (header, data)


def loadTrace(fname, tryCompressedIfMissing=True):
    """Load trace file."""
    (header, data) = loadTraceRawHeaderAndData(
        fname, tryCompressedIfMissing=tryCompressedIfMissing
    )
    return traceFromHeaderAndData(header, data)


def traceFromHeaderAndData(header, data=None):
    """Generate trace class from it description (header) and data."""
    label = None
    model = None
    tr = None
    if "config" not in header:
        print("Error: trace has now config")
        return None
    else:
        if "label" in header["config"]:
            label = header["config"]["label"]
        if "model" not in header["config"]:
            print("Error: unknown trace model")
            return None
        else:
            model = header["config"]["model"]
            if model == "Trace":
                tr = Trace(label)
                if data is not None:
                    tr.values = data
            elif model == "TraceXY":
                tx = traceFromHeaderAndData(header["TraceX"])
                ty = traceFromHeaderAndData(header["TraceY"])
                if data is not None:
                    tx.values = data[:, 0]
                    ty.values = data[:, 1]
                tr = TraceXY(label)
                tr.x = tx
                tr.y = ty
            elif model == "TraceSetSameX":
                tx = traceFromHeaderAndData(header["TraceX"])
                tx.values = data[:, 0]
                tr = TraceSetSameX(label)
                tr.addTraceX(tx)
                ytrs_header = header["TraceY"]
                cnt = 0
                for name, h in ytrs_header.items():
                    ty = traceFromHeaderAndData(h)
                    cnt += 1
                    ty.values = data[:, cnt]
                    trxy = TraceXY(name)
                    trxy.x = tx
                    trxy.y = ty
                    tr.addTrace(trxy)

            else:
                print(f"Error: unknown trace model: {model}")
                return None
        tr.config = header["config"]

    return tr


class Trace:
    """Base Trace class, which holds only one variable"""

    def __init__(self, label):
        self.config = {}
        self.config["label"] = label
        self.config["model"] = "Trace"
        self.config["version"] = "0.1"
        # 'type' is useful to indicate way of representation,
        # it makes sense on y_vs_x traces.
        # If set to none we have normal y vs x.
        # If set to 'timestamp' x will be converted to datetime dates
        self.config["type"] = None
        self.config["item_format"] = ".15e"
        self.config["tags"] = {}
        self.last_saved_pos = 0
        self._trace_specific_init()
        self.clear_data()

    def _trace_specific_init(self):
        self.config["unit"] = None
        self.values = np.empty(0)
        self.last_saved_pos = 0

    def clear_last_saved_pos(self):
        self.last_saved_pos = 0

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.values is not None:
            self.values = np.empty(0, dtype=self.values.dtype)

    def __repr__(self):
        lbl = self.config["label"]
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        return "".join([cls_name, f", N={self.values.size}", ")"])

    def plot(self):
        import matplotlib.pyplot as plt

        x = self.values
        if self.config["type"] is not None:
            if self.config["type"] == "timestamp":
                x = from_timestamps_to_dates(self.values)
        plt.plot(x, label=self.config["label"])
        plt.xlabel("index")
        plt.ylabel(f"{self.config['unit']}")
        plt.legend()
        plt.grid(True)

    def getConfig(self):
        d = {}
        d["config"] = self.config.copy()
        return d

    def getData(self):
        return self.values

    def getHeader(self, prefix=""):
        d = self.getConfig()
        return headerFromDictionary(d, prefix="")

    def save(
        self, fname, last_saved_pos=None, skip_headers_if_file_exist=False, **kwargs
    ):
        """Save trace to file and returns its filename."""
        if last_saved_pos is None:
            last_saved_pos = self.last_saved_pos
        data = self.getData()
        if last_saved_pos > 0:
            skip_headers_if_file_exist = True
        fname = save_table_with_header(
            fname,
            data[last_saved_pos:, :],
            self.getHeader(),
            item_format=self.config["item_format"],
            skip_headers_if_file_exist=skip_headers_if_file_exist,
            **kwargs,
        )
        self.last_saved_pos = data.shape[0]
        return fname

    def addPoint(self, val):
        self.values = np.append(self.values, val)


class TraceXY(Trace):
    """Data structure to handle two linked variables X and Y.

    It is handy for Y vs X data arrangements.
    """

    def __init__(self, label):
        super().__init__(label)
        self.config["model"] = "TraceXY"

    def _trace_specific_init(self):
        self.x = None
        self.y = None

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.x is not None:
            self.x.clear_data()
        if self.y is not None:
            self.y.clear_data()

    def __repr__(self):
        lbl = self.config["label"]
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        # xlabel = f"{self.x.config['label']}"
        xparam = f", {self.x}"
        yparam = f", {self.y}"
        return "".join([cls_name, xparam, yparam, ")"])

    def plot(self):
        import matplotlib.pyplot as plt

        x = self.x.values
        if self.x.config["type"] is not None:
            if self.x.config["type"] == "timestamp":
                x = from_timestamps_to_dates(x)
        plt.plot(x, self.y.values, label=self.config["label"])
        plt.xlabel(f"{self.x.config['label']} ({self.x.config['unit']})")
        plt.ylabel(f"{self.y.config['label']} ({self.y.config['unit']})")
        plt.legend()
        plt.grid(True)

    def getConfig(self):
        config = {}
        config["config"] = self.config.copy()
        config["TraceX"] = {}
        config["TraceX"] = self.x.getConfig()
        config["TraceY"] = {}
        config["TraceY"] = self.y.getConfig()
        return config

    def getData(self):
        data = self.x.values
        if data.ndim == 1:
            data = data[:, np.newaxis]
        vals = self.y.values
        if vals.ndim == 1:
            vals = vals[:, np.newaxis]
        data = np.concatenate((data, vals), 1)
        return data

    def addPoint(self, valX, valY):
        self.x.values = np.append(self.x.values, valX)
        self.y.values = np.append(self.y.values, valY)


class TraceSetSameX(Trace):
    """Data structure to handle multiple Ys vs X dependencies.
    It is handy for scope traces."""

    def __init__(self, label):
        super().__init__(label)
        self.config["model"] = "TraceSetSameX"

    def _trace_specific_init(self):
        self.x = None
        self.traces = {}

    def clear_data(self):
        self.clear_last_saved_pos()
        if self.x is not None:
            self.x.clear_data()
        for k, tr in self.traces.items():
            tr.clear_data()

    def __repr__(self):
        lbl = self.config["label"]
        cls_name = f"{self.__class__.__name__}('{lbl}'"
        xparam = f", x: {self.x}"
        yparam = f", traces: {list(self.traces.keys())}"
        return "".join([cls_name, xparam, yparam, ")"])

    def addTraceX(self, tr):
        self.x = tr

    def addTrace(self, tr):
        if tr.config["model"] == "TraceXY":
            if self.x is None:
                self.x = tr.x
            trY = tr.y
            self.traces[tr.config["label"]] = trY
        elif tr.config["model"] == "Trace":
            if self.x is None:
                self.x = tr
            else:
                self.traces[tr.config["label"]] = tr

    def plot(self):
        import matplotlib.pyplot as plt

        nplots = len(self.traces.keys())
        fig = plt.gcf()
        fig, axs = plt.subplots(nplots, 1, sharex=True, num=fig.number)
        cnt = 0
        x = self.x.values
        if self.x.config["type"] is not None:
            if self.x.config["type"] == "timestamp":
                x = from_timestamps_to_dates(x)
        for k, tr in self.traces.items():
            axs[cnt].plot(x, tr.values, label=k)
            axs[cnt].set_ylabel(f"{tr.config['label']} ({tr.config['unit']})")
            axs[cnt].legend()
            axs[cnt].grid(True)
            cnt += 1
        plt.xlabel(f"{self.x.config['label']} ({self.x.config['unit']})")

    def items(self):
        return self.traces.items()

    def keys(self):
        return self.traces.keys()

    def getTrace(self, label):
        tr = TraceXY(label)
        tr.x = self.x
        tr.y = self.traces[label]
        return tr

    def getConfig(self):
        config = {}
        config["config"] = self.config.copy()
        config["TraceX"] = {}
        config["TraceX"] = self.x.getConfig()
        config["TraceY"] = {}
        for k, v in self.traces.items():
            config["TraceY"][k] = v.getConfig()
        return config

    def getData(self):
        data = self.x.values
        if data.ndim == 1:
            data = data[:, np.newaxis]
        for k, v in self.traces.items():
            vals = v.values
            if vals.ndim == 1:
                vals = vals[:, np.newaxis]
            data = np.concatenate((data, vals), 1)
        return data

    def addPointToTrace(self, val, name=None):
        if name is None:
            self.x.values = np.append(self.x.values, val)
        else:
            a = self.traces[name].values
            a = np.append(a, val)
            self.traces[name].values = a


if __name__ == "__main__":
    print("Testing trace")
    x = Trace("x trace")
    x.values = np.random.normal(2, 2, (4, 1))
    x.values = np.array(x.values, int)
    x.config["unit"] = "s"
    x.config["tags"]["tag1"] = "xxxx"
    x.config["tags"]["tag2"] = "xxxx"
    x.save("xtrace.dat", skip_headers_if_file_exist=True)
    # print(x.getHeader())
    y = Trace("y trace")
    y.values = np.random.normal(2, 2, (4, 1))
    y.config["unit"] = "V"
    y.config["tags"]["ytag2"] = "yyyy"
    xy = TraceXY("xy trace")
    xy.config["tags"]["xy tag"] = "I am xy tag"
    xy.x = x
    xy.y = y
    xy.save("xytrace.dat")
    # print(xy.getHeader())
    xyn = TraceSetSameX("many ys trace")
    xyn.config["tags"]["descr"] = "I am many ys trace"
    xy.config["label"] = "y1"
    xyn.addTrace(xy)
    xy.config["label"] = "y2"
    xyn.addTrace(xy)
    xy.config["label"] = "y3"
    xyn.addTrace(xy)
    xyn.save("xyntrace.dat")
    # print(xyn.getHeader())



"""Basic class to build hardware avare classes of power supplies."""

from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument


class PowerSupply(BasicInstrument):
    """Base class for a Power Supply.

    Contains minimal set of methods to be implemented by a Power Supply.

    Intended to be used as a parent for hardware aware power supplies.
    """

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "PowerSupply"
        self.config["Device model"] = (
            "Generic Power Supply generator Without Hardware interface"
        )
        self.config["FnamePrefix"] = "power_supply"
        self.deviceProperties.update({})


class PowerSupplySCPI(SCPIinstr, PowerSupply):
    """SCPI aware power supply.

    Intended to be used as a parent for hardware aware power supplies.

    Example
    -------

    >>> rm = pyvisa.ResourceManager()
    >>> PowerSupplySCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))

    or

    >>> PowerSupplySCPI(rm.open_resource('USB0::10893::4354::MY61001869::0::INSTR'))
    """

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        PowerSupply.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



"""
Module: HP 3457A Digital Multimeter Control
Description: This module provides a class to interface with a digital multimeter (DMM)
             using pyvisa for performing measurements such as voltage, current,
             and resistance.

Author: Mykhailo Vorobiov
Email: mvorobiov@wm.edu
Date: 2024-10-02
Updated: 2024-11-05
"""

from qolab.hardware.basic import BasicInstrument
from ._basic import Multimeter

import pyvisa
import numpy as np

class HP3457A(Multimeter):
    """
    A class to interface with HP 3457A Digital Multimeter (DMM) via pyvisa.

    This class allows you to measure DC voltage, AC voltage, DC current, AC current,
    and resistance from a DMM using pyvisa communication.
    """

    # List of data return formats
    FORMATS = ['ASCII', 
               'SINT', 
               'DINT', 
               'SREAL', 
               'DREAL'
               ]
    
    BEEPER_STATUS = ['ON','OFF','ONCE']

    # List of functions
    FUNCTIONS = ['DCV',
                 'ACV',
                 'ACDCV',#(This is here for future work)
                 'OHM',#(This is here for future work)
                 'OHMF',#(This is here for future work)
                 'DCI',
                 'ACI',
                 'ACDCI',#(This is here for future work)
                 'FREQ',#(This is here for future work)
                 'PER'#(This is here for future work)
                ]
    
    # Allowed number of power line cycles (defines A/D integration time)
    NPLC = [0.0005, 0.005, .1, 1, 10, 100]

    # DC Volts Spec (This is here for future work)
    DCV_RES = {'30mv': {'6.5': 10e-9, '5.5': 100e-9, '4.5': 1e-6, '3.5': 10e-6},
                    '300mv': {'6.5': 100e-9, '5.5': 1e-6, '4.5': 10e-6, '3.5': 100e-6},
                    '3v': {'6.5': 1e-6, '5.5': 10e-6, '4.5': 100e-6, '3.5': 1e-3},
                    '30v': {'6.5': 10e-6, '5.5': 100e-6, '4.5': 1e-3, '3.5': 10e-3},
                    '300v': {'6.5': 100e-6, '5.5': 1e-3, '4.5': 10e-3, '3.5': 100e-3}
                    }
    
    DCV_ACC = {'30mv': {'100': {'acc': .0045, 'counts': 365}, '10': {'acc': .0045, 'counts': 385},
                    '1': {'acc': .0045, 'counts': 500}, '.1': {'acc': .0045, 'counts': 70},
                    '.005': {'acc': .0045, 'counts': 19}, '.0005': {'acc': .0045, 'counts': 6}},
                    '300mv': {'100': {'acc': .0035, 'counts': 39}, '10': {'acc': .0035, 'counts': 40},
                    '1': {'acc': .0035, 'counts': 50}, '.1': {'acc': .0035, 'counts': 9},
                    '.005': {'acc': .0035, 'counts': 4}, '.0005': {'acc': .0035, 'counts': 4}},
                    '3v': {'100': {'acc': .0025, 'counts': 6}, '10': {'acc': .0025, 'counts': 7},
                    '1': {'acc': .0025, 'counts': 7}, '.1': {'acc': .0025, 'counts': 4},
                    '.005': {'acc': .0025, 'counts': 4}, '.0005': {'acc': .0025, 'counts': 4}},
                    '30v': {'100': {'acc': .0040, 'counts': 19}, '10': {'acc': .0040, 'counts': 20},
                    '1': {'acc': .0040, 'counts': 30}, '.1': {'acc': .0040, 'counts': 7},
                    '.005': {'acc': .0040, 'counts': 4}, '.0005': {'acc': .0040, 'counts': 4}},
                    '300v': {'100': {'acc': .0055, 'counts': 6}, '10': {'acc': .0055, 'counts': 7},
                    '1': {'acc': .0055, 'counts': 7}, '.1': {'acc': .0055, 'counts': 4},
                    '.005': {'acc': .0055, 'counts': 4}, '.0005': {'acc': .0055, 'counts': 4}}
                    }
    
    # DC Current Spec (This is here for future work)
    DCI_RES = {'300ua': {'6.5': 100e-12, '5.5': 1e-9, '4.5': 10e-9, '3.5': 100e-9},
                    '3ma': {'6.5': 1e-9, '5.5': 10e-9, '4.5': 100e-9, '3.5': 1e-6},
                    '30ma': {'6.5': 10e-9, '5.5': 100e-9, '4.5': 1e-6, '3.5': 10e-6},
                    '300ma': {'6.5': 100e-9, '5.5': 1e-6, '4.5': 10e-6, '3.5': 100e-6},
                    '1a': {'6.5': 1e-6, '5.5': 10e-6, '4.5': 100e-6, '3.5': 1e-3}
                    }
    # (This is here for future work)
    DCI_ACC = {'300ua': {'100': {'acc': .04, 'counts': 104}, '10': {'acc': .04, 'counts': 104},
                    '1': {'acc': .04, 'counts': 115}, '.1': {'acc': .04, 'counts': 14},
                    '.005': {'acc': .04, 'counts': 5}, '.0005': {'acc': .04, 'counts': 4}},
                    '3ma': {'100': {'acc': .04, 'counts': 104}, '10': {'acc': .04, 'counts': 104},
                    '1': {'acc': .04, 'counts': 115}, '.1': {'acc': .04, 'counts': 14},
                    '.005': {'acc': .04, 'counts': 5}, '.0005': {'acc': .04, 'counts': 4}},
                    '30ma': {'100': {'acc': .04, 'counts': 104}, '10': {'acc': .04, 'counts': 104},
                    '1': {'acc': .04, 'counts': 115}, '.1': {'acc': .04, 'counts': 14},
                    '.005': {'acc': .04, 'counts': 5}, '.0005': {'acc': .04, 'counts': 4}},
                    '300ma': {'100': {'acc': .08, 'counts': 204}, '10': {'acc': .08, 'counts': 204},
                    '1': {'acc': .08, 'counts': 215}, '.1': {'acc': .08, 'counts': 24},
                    '.005': {'acc': .08, 'counts': 6}, '.0005': {'acc': .08, 'counts': 4}},
                    '1a': {'100': {'acc': .08, 'counts': 604}, '10': {'acc': .08, 'counts': 604},
                    '1': {'acc': .08, 'counts': 615}, '.1': {'acc': .08, 'counts': 64},
                    '.005': {'acc': .08, 'counts': 10}, '.0005': {'acc': .08, 'counts': 5}}
                    }

    def __init__(self, resource, *args, **kwds):
        """
        Initialize the Digital Multimeter class and establish a connection.

        :param resource_name: VISA resource name for the multimeter.
        :param alias: Alias for logging.
        :param log_level: Logging level (default: INFO).
        """
        super().__init__(*args, **kwds)
        self.resource = resource
        self.config["Device model"] = "HP 3457A"
        self.resource.timeout = 5000
        self.resource.read_termination = '\r\n'
        self.resource.write_termination = '\r\n'

        self.read = self.resource.read
        self.write = self.resource.write
        self.read_bytes = self.resource.read_bytes

        self.switchTime = 0.5  # switch time in seconds for Function/Measurement change
        self.deviceProperties.update({"Function"})
        self.set_function('DCV')


    def get_idn(self):
        """
        Query the identification of the digital multimeter: manufacturer, model, serial number, and firmware version.

        :return: Identifier string if query is successful, or None otherwise.
        """
        try:
            self.write('ID?')
            return self.read_bytes(9).decode().strip()
        except pyvisa.VisaIOError as e:
            print(f'[ERROR!] Error retrieving identification of the digital multimeter: {e}')
            return None
    
    def _check_format(self, format):
        if format in self.FORMATS:
            return True
        else:
            return False
        
    def set_format(self, format='ASCII'):
        if self._check_format(format):
            try:
                self.write(f'OFORMAT {format}')
                print(f'[INFO] Format  has been set to {format}')
            except pyvisa.VisaIOError as e:
                print(f'[ERROR!] Couldn\'t set reading format to {format}: {e}')
        else:
            print(f'[WARNING] Entered format is not allowed. Allowed formats: {self.FORMATS}')
    
    @BasicInstrument.tsdb_append
    def get_reading(self):
        try:
            self.write(f'TARM AUTO')
            return float(self.read_bytes(16).decode().strip())
        except pyvisa.VisaIOError as e:
            print(f'[ERROR!] Error failed to get reading: {e}')
            return np.nan()

    def set_beeper_status(self, status='OFF'):
        if status in self.BEEPER_STATUS:
            try:
                self.write(f'BEEP {status}')
                print(f'[INFO] Beeper is set to {status}')
            except pyvisa.VisaIOError as e:
                print(f'[ERROR!] Error setting beeper status: {e}')
        else:
            print(f'[WARNING!] The passed beeper status is not allowed.'
                                f'Choose from {self.BEEPER_STATUS}')

    def set_function(self, function='DCV'):
        """
        Set the measurement function of the multimeter.

        :param function: Measurement function .
        """
        if function in self.FUNCTIONS:
            try:
                self.write(f"FUNC {function}")
                print(f"[INFO] Measurement function set to {function}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting measurement function: {e}")
        else:
            print(f"[WARNING!] Invalid measurement function '{function}'. "
                                f"Valid functions are: {', '.join(self.FUNCTIONS)}")

    @BasicInstrument.tsdb_append
    def get_temperature(self):
        try:
            self.write("TEMP?")
            temp = float(self.read(16).decode().strip())
            print(f"[INFO] DMM's internal temperature is: {temp}")
            return temp
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error retrieving DMM's internal temperature: {e}")

    def get_memory_count(self):
        """
        Returns the total number of stored readings.
        """
        try:
            self.write(f'MCOUNT?')
            return float(self.read_bytes(16).decode().strip())
        except pyvisa.VisaIOError as e:
            print(f'[ERROR!] Error failed to get count of reading in memory: {e}')
            return np.nan()

    def enable_keyboard(self, status=True):
        if status:
            try:
                self.write("LOCK ON")
                print(f"[INFO] DMM's keyboard is unlocked.")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error unlocking the DMM's keyboard: {e}")
        else:
            try:
                self.write("LOCK OFF")
                print(f"[INFO] DMM's keyboard is locked.")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error locking the DMM's keyboard: {e}")

    def set_nplc(self, nplc=10):
        if nplc in self.NPLC:
            try:
                self.write(f"FUNC {nplc}")
                print(f"[INFO] NPLC is set to {nplc}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting number of power line cycles (NPLC): {e}")
        else:
            print(f"[WARNING!] Invalid NPLC '{nplc}'. "
                                f"Valid NPLC values are: {', '.join(str(self.NPLC))}")

        
    #-----------------------------------------------------------
    # Methods to comply with other DMM QOLab module drivers
    #-----------------------------------------------------------
    @BasicInstrument.tsdb_append
    def getVdc(self):
        self.set_function('DCV')
        return self.get_reading()
    
    @BasicInstrument.tsdb_append
    def getVac(self):
        self.set_function('ACV')
        return self.get_reading()
    
    @BasicInstrument.tsdb_append
    def getAdc(self):
        self.set_function('DCI')
        return self.get_reading()
    
    @BasicInstrument.tsdb_append
    def getAac(self):
        self.set_function('ACI')
        return self.get_reading()

# Example usage
if __name__ == "__main__":
    import time

    rm = pyvisa.ResourceManager()
    instrument = rm.open_resource("visa://192.168.194.15/GPIB1::22::INSTR")
    dmm = HP3457A(instrument)
    dmm.set_format('ASCII')
    dmm.set_beeper_status('ONCE')
    dmm.set_function('DCI')
    print(dmm.get_reading())
    dmm.close()



from qolab.hardware.scpi import response2numStr


def test_separator_and_unit():
    assert response2numStr("TDIV 2.00E-08S", firstSeparator=" ", unit="Z") == (
        "TDIV",
        "2.00E-08S",
        None,
    )  # incorrect unit was specified
    assert response2numStr("TDIV 2.00E-08S", firstSeparator=" ", unit="S") == (
        "TDIV",
        "2.00E-08",
        "S",
    )
    assert response2numStr("TDIV,2.00E-08S", firstSeparator=",", unit="S") == (
        "TDIV",
        "2.00E-08",
        "S",
    )


def test_separator_and_empty_unit():
    assert response2numStr("TDIV,2.00E-08", firstSeparator=",", unit=None) == (
        "TDIV",
        "2.00E-08",
        None,
    )
    assert response2numStr("TDIV,2.00E-08", firstSeparator=",", unit="") == (
        "TDIV",
        "2.00E-08",
        "",
    )


def test_no_separator_with_unit():
    assert response2numStr("2.00E-08S", firstSeparator=None, unit="S") == (
        None,
        "2.00E-08",
        "S",
    )
    assert response2numStr("2.00E-08S", firstSeparator="", unit="S") == (
        None,
        "2.00E-08",
        "S",
    )


def test_no_separator_and_no_unit():
    assert response2numStr("2.00E-08", firstSeparator=None, unit=None) == (
        None,
        "2.00E-08",
        None,
    )
    assert response2numStr("2.00E-08", firstSeparator="", unit="") == (
        None,
        "2.00E-08",
        "",
    )
